# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Sequence

from xpra.os_util import WIN32, OSX, gi_import
from xpra.common import roundup
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.util.env import envbool, first_time
from xpra.codecs.constants import VideoSpec, get_profile
from xpra.gstreamer.common import (
    import_gst, normv, get_all_plugin_names,
    get_caps_str, get_element_str, wrap_buffer,
    get_default_appsink_attributes, get_default_appsrc_attributes,
    BUFFER_FORMAT, GST_FLOW_OK,
)
from xpra.codecs.gstreamer.common import (
    VideoPipeline,
    get_version, get_type, get_info,
    get_gst_encoding, get_gst_rgb_format, get_video_encoder_caps,
    get_video_encoder_options,
)
from xpra.codecs.image import ImageWrapper
from xpra.log import Logger

GObject = gi_import("GObject")
Gst = import_gst()
log = Logger("encoder", "gstreamer")

NVIDIA_VAAPI = envbool("XPRA_NVIDIA_VAAPI", False)
VAAPI = envbool("XPRA_GSTREAMER_VAAPI", not (WIN32 or OSX))
NVENC = envbool("XPRA_GSTREAMER_NVENC", False)
NVD3D11 = envbool("XPRA_GSTREAMER_NVD3D11", WIN32)
FORMATS = os.environ.get("XPRA_GSTREAMER_ENCODER_FORMATS", "h264,hevc,vp8,vp9,av1").split(",")

log(f"encoder: {get_type()} {get_version()}")

PACKED_RGB_FORMATS = ("RGBA", "BGRA", "ARGB", "ABGR", "RGB", "BGR", "BGRX", "XRGB", "XBGR")

assert get_type()  # all codecs must define this function
COLORSPACES: dict[str, dict[str, list[str]]] = {}


def get_encodings() -> Sequence[str]:
    return tuple(COLORSPACES.keys())


def ElementEncoderClass(element: str):
    class ElementEncoder(Encoder):
        pass

    ElementEncoder.encoder_element = element
    return ElementEncoder


def make_spec(element: str, encoding: str, cs_in: str, css_out: Sequence[str],
              cpu_cost: int = 50, gpu_cost: int = 50) -> VideoSpec:
    # use a metaclass so all encoders are gstreamer.encoder.Encoder subclasses,
    # each with different pipeline arguments based on the make_spec parameters:
    if cs_in in PACKED_RGB_FORMATS:
        width_mask = height_mask = 0xFFFF
    else:
        width_mask = height_mask = 0xFFFE
    spec = VideoSpec(
        encoding=encoding, input_colorspace=cs_in,
        output_colorspaces=css_out,
        has_lossless_mode=False,
        codec_class=ElementEncoderClass(element), codec_type=f"gstreamer-{element}",
        quality=40, speed=40,
        setup_cost=100, cpu_cost=cpu_cost, gpu_cost=gpu_cost,
        width_mask=width_mask, height_mask=height_mask,
        min_w=64, min_h=64,
        max_w=4096, max_h=4096)
    spec.gstreamer_element = element
    return spec


SPECS: dict[str, dict[str, list[VideoSpec]]] = {}


def get_specs() -> Sequence[VideoSpec]:
    # easy, just flatten the ones we found:
    specs: Sequence[VideoSpec] = []
    for cs_dicts in SPECS.values():
        for lspecs in cs_dicts.values():
            specs += lspecs
    return specs


def init_all_specs(*exclude) -> None:
    # by default, try to enable everything
    # the self-tests should disable what isn't available / doesn't work
    specs: dict[str, dict[str, list]] = {}
    colorspaces: dict[str, dict[str, list]] = {}
    missing: list[str] = []

    def add(element: str, encoding: str, cs_in: str, css_out, *args):
        if element in missing:
            return
        if encoding not in FORMATS:
            log(f"{encoding} not enabled - supported formats: {FORMATS}")
            return
        if element in exclude:
            return
        if element not in get_all_plugin_names():
            if element not in ("amfh264enc", "amfh265enc"):
                missing.append(element)
            return
        # add spec:
        css_out = css_out or (cs_in,)
        spec = make_spec(element, encoding, cs_in, css_out, *args)
        specs.setdefault(encoding, {}).setdefault(cs_in, []).append(spec)
        # update colorspaces map (add new output colorspaces - if any):
        cur = colorspaces.setdefault(encoding, {}).setdefault(cs_in, [])
        for v in css_out:
            if v not in cur:
                cur.append(v)

    vaapi = VAAPI
    if VAAPI and not NVIDIA_VAAPI:
        try:
            from xpra.codecs.nvidia.util import has_nvidia_hardware
            vaapi = not has_nvidia_hardware()
        except ImportError:
            pass
    log(f"init_all_specs try vaapi? {vaapi}")
    if vaapi:
        add("vah264lpenc", "h264", "NV12", ("YUV420P",), 20, 100)
        add("vah264enc", "h264", "NV12", ("YUV420P",), 20, 100)
        add("vaapih264enc", "h264", "NV12", ("YUV420P",), 20, 100)
        add("vaapih265enc", "hevc", "NV12", ("YUV420P",), 20, 100)
    if NVD3D11:
        add("nvd3d11h264enc", "h264", "YUV420P", ("YUV420P",), 20, 100)
        add("nvd3d11h265enc", "hevc", "YUV420P", ("YUV420P",), 20, 100)
    if NVENC:
        add("nvh264enc", "h264", "YUV420P", ("YUV420P",), 20, 100)
        add("nvh265enc", "hevc", "YUV420P", ("YUV420P",), 20, 100)
    if not (OSX or WIN32):
        add("amfh264enc", "h264", "NV12", ("YUV420P",), 20, 100)
        add("amfh265enc", "hevc", "NV12", ("YUV420P",), 20, 100)
    add("x264enc", "h264", "YUV420P", ("YUV420P",), 100, 0)
    add("x264enc", "h264", "YUV444P", ("YUV444P",), 100, 0)
    add("openh264enc", "h264", "YUV420P", ("YUV420P",), 100, 0)
    add("vp8enc", "vp8", "YUV420P", ("YUV420P",), 100, 0)
    add("vp9enc", "vp9", "YUV444P", ("YUV444P",), 100, 0)
    if not OSX:
        add("av1enc", "av1", "YUV420P", ("YUV420P",), 100, 0)
        add("av1enc", "av1", "YUV444P", ("YUV444P",), 100, 0)
    # svt encoders error out:
    # add("svtav1enc", "av1", "YUV420P", ("YUV420P", ), 100, 0)
    # add("svtvp9enc", "vp9", "YUV420P", ("YUV420P", ), 100, 0)
    # add: nvh264enc, nvh265enc ?
    global SPECS, COLORSPACES
    SPECS = specs
    COLORSPACES = colorspaces
    log("init_all_specs%s SPECS=%s", exclude, SPECS)
    log("init_all_specs%s COLORSPACES=%s", exclude, COLORSPACES)
    if missing and first_time("gstreamer-encoder-missing-elements"):
        log.info("some GStreamer elements are missing or unavailable on this system:")
        log.info(" " + csv(missing))


class Encoder(VideoPipeline):
    __gsignals__ = VideoPipeline.__generic_signals__.copy()
    encoder_element = "unset"

    def __repr__(self):
        if self.colorspace is None:
            return f"gstreamer-{self.encoder_element}(uninitialized)"
        return f"gstreamer-{self.encoder_element}({self.colorspace} - {self.width}x{self.height})"

    """
    Dispatch video encoding to a gstreamer pipeline
    """

    def create_pipeline(self, options: typedict) -> None:
        if self.encoding not in get_encodings():
            raise ValueError(f"invalid encoding {self.encoding!r}")
        self.dst_formats = options.strtupleget("dst-formats")
        gst_rgb_format = get_gst_rgb_format(self.colorspace)
        if not gst_rgb_format:
            raise ValueError(f"unable to map {self.colorspace} to a gstreamer pixel format")
        vcaps: dict[str, Any] = {
            "width": self.width,
            "height": self.height,
            "format": gst_rgb_format,
            "framerate": (60, 1),
            "interlace": "progressive",
            "colorimetry": "bt709",
        }
        CAPS = get_caps_str("video/x-raw", vcaps)
        self.profile = self.get_profile(options)  # ie: "high"
        eopts = get_video_encoder_options(self.encoder_element, self.profile, options)
        vcaps = get_video_encoder_caps(self.encoder_element)
        self.extra_client_info = vcaps.copy()
        if self.profile:
            vcaps["profile"] = self.profile
            self.extra_client_info["profile"] = self.profile
        appsrc_opts = get_default_appsrc_attributes()
        appsrc_opts |= {
            "is-live": True,
            "do-timestamp": True,
            "format": BUFFER_FORMAT,
            "caps": CAPS,
            # "leaky-type"    : 0,        # default is 0 and this is not available before GStreamer 1.20
        }
        gst_encoding = get_gst_encoding(self.encoding)  # ie: "hevc" -> "video/x-h265"
        elements = [
            get_element_str("appsrc", appsrc_opts),
            get_element_str(self.encoder_element, eopts),
            get_caps_str(gst_encoding, vcaps),
            get_element_str("appsink", get_default_appsink_attributes())
        ]
        if not self.setup_pipeline_and_bus(elements):
            raise RuntimeError("failed to setup gstreamer pipeline")

    def get_profile(self, options: typedict) -> str:
        default_profile: str = {
            # "x264enc"   : "constrained-baseline",
            # "vaapih264enc" : "constrained-baseline",
            # "nvh264enc" : "main",
            "vp8enc": "",  # 0-4
            "vp9enc": "",  # 0-4
        }.get(self.encoder_element, "")
        return get_profile(options, self.encoding, self.colorspace, default_profile)

    def get_src_format(self) -> str:
        return self.colorspace

    def get_info(self) -> dict[str, Any]:
        info = super().get_info()
        if self.dst_formats:
            info["dst_formats"] = self.dst_formats
        return info

    def clean(self) -> None:
        self.cleanup()

    def on_new_sample(self, _bus) -> int:
        sample = self.sink.emit("pull-sample")
        buf = sample.get_buffer()
        size = buf.get_size()
        log("on_new_sample size=%s", size)
        if size:
            data = buf.extract_dup(0, size)
            client_info = self.extra_client_info
            self.extra_client_info = {}
            client_info["frame"] = self.frames
            self.frames += 1
            pts = normv(buf.pts)
            if pts >= 0:
                client_info["timestamp"] = pts
            duration = normv(buf.duration)
            if duration >= 0:
                client_info["duration"] = duration
            qs = self.frame_queue.qsize()
            if qs > 0:
                client_info["delayed"] = qs
            self.frame_queue.put((data, client_info))
            log(f"added data to frame queue, client_info={client_info}")
        return GST_FLOW_OK

    def compress_image(self, image: ImageWrapper, options: typedict):
        if image.get_planes() == ImageWrapper.PACKED:
            data = image.get_pixels()
            rowstride = image.get_rowstride()
            want_rowstride = roundup(self.width, 2) * len(self.colorspace)
            if rowstride != want_rowstride and not image.restride(want_rowstride):
                raise RuntimeError(f"failed to restride image from {rowstride}to {want_rowstride}")
        else:
            # merge all planes into a single buffer:
            data = b"".join(image.get_pixels())
        log(f"compress_image({image}, {options}) state={self.state} pixel buffer size={len(data)}")
        if self.state in ("stopped", "error"):
            log(f"pipeline is in {self.state} state, dropping buffer")
            return None
        return self.process_buffer(wrap_buffer(data), options)


GObject.type_register(Encoder)


def selftest(_full=False):
    log("gstreamer encoder selftest: %s", get_info())
    from xpra.codecs.checks import test_encoder_spec, DEFAULT_TEST_SIZE
    W, H = DEFAULT_TEST_SIZE
    # test individual specs
    skip = []
    log(f"will self test: {SPECS}")
    for encoding, cs_map in SPECS.items():
        for cs_in, specs in cs_map.items():
            for spec in specs:
                try:
                    for cs_out in spec.output_colorspaces:
                        test_encoder_spec(spec.codec_class, encoding, cs_in, cs_out, W, H)
                        log(f"{spec.gstreamer_element} {encoding} {cs_in} -> {spec.output_colorspaces} passed")
                except Exception as e:
                    log("test_encoder_spec", exc_info=True)
                    log.warn(f"Warning: gstreamer {spec.gstreamer_element!r} encoder failed")
                    log.warn(f" {e}")
                    skip.append(spec.gstreamer_element)
    init_all_specs(*skip)


init_all_specs()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time
import uuid
import unittest

from xpra.util.env import envint
from xpra.exit_codes import exit_str
from xpra.os_util import OSX, POSIX
from xpra.util.io import load_binary_file, pollwait
from xpra.platform.paths import get_download_dir
from unit.client.x11_client_test_util import X11ClientTestUtil, log

CLIENT_TIMEOUT = envint("XPRA_TEST_CLIENT_TIMEOUT", 20)


class X11ClientTest(X11ClientTestUtil):

    def do_test_connect(self, disconnect=True, client_args=(), server_args=()):
        display = self.find_free_display()
        log("starting test server on %s", display)
        server_args = ["--start=xterm"] + list(server_args)
        server = self.check_fast_start_server(display, *server_args)
        xvfb1, client1 = self.run_client(display, *client_args)
        r = pollwait(client1, CLIENT_TIMEOUT)
        if r is not None:
            raise RuntimeError(f"client1 with args {client_args} exited with code {exit_str(r)}")
        xvfb2, client2 = self.run_client(display, *client_args)
        r = pollwait(client2, CLIENT_TIMEOUT)
        if r is not None:
            raise RuntimeError(f"client2 with args {client_args} exited with code {exit_str(r)}")
        if disconnect:
            #starting a second client should disconnect the first when not sharing
            assert pollwait(client1, 2) is not None, "the first client should have been disconnected"
        #killing the Xvfb should kill the client
        xvfb1.terminate()
        xvfb2.terminate()
        assert pollwait(xvfb1, CLIENT_TIMEOUT) is not None, "xvfb1 has not terminated"
        assert pollwait(xvfb2, CLIENT_TIMEOUT) is not None, "xvfb2 has not terminated"
        assert pollwait(client1, CLIENT_TIMEOUT) is not None, "client1 has not terminated"
        assert pollwait(client2, CLIENT_TIMEOUT) is not None, "client2 has not terminated"
        server.terminate()


    def test_connect(self):
        self.do_test_connect()

    def test_sharing(self):
        self.do_test_connect(False, server_args=("--sharing=yes",))

    def test_opengl(self):
        self.do_test_connect(client_args=("--opengl=yes",))

    def test_multiscreen(self):
        client_display = self.find_free_display()
        xvfb = self.start_Xvfb(client_display, screens=[(1024,768), (1200, 1024)])
        #multiscreen requires Xvfb, which may not support opengl:
        self.do_run_client(client_display, "--opengl=no").terminate()
        xvfb.terminate()

    def test_nocomposite(self):
        client_display = self.find_free_display()
        self.start_Xvfb(client_display, extensions=("-Composite"))
        self.do_run_client(client_display).terminate()


    def _test_client_depth(self, server_display):
        #start vfb with display-depth:
        for client_display_depth in (16, 24, 30):
            client_display = self.find_free_display()
            xvfb = self.start_Xvfb(client_display, depth=client_display_depth)
            try:
                for client_depth in (16, 24, 30):
                    client = self.do_run_client(client_display, server_display,
                                            "--pixel-depth=%i" % client_depth)
                    r = pollwait(client, 5)
                    assert r is None, "client has terminated with exit code %i" % r
                    client.terminate()
            finally:
                xvfb.terminate()

    def test_depth(self):
        for server_depth in (16, 24, 30):
            server_display = self.find_free_display()
            log("depth=%i starting test server on %s", server_depth, server_display)
            server = self.check_fast_start_server(server_display,
                                            "--start=xterm", "--sync-xvfb=50",
                                            "--pixel-depth=%i" % server_depth)
            self._test_client_depth(server_display)
            server.terminate()

    def do_test_control_send_file(self, data):
        f = self._temp_file(data)
        client = xvfb = server = None
        try:
            display = self.find_free_display()
            server = self.check_fast_start_server(display, "--file-transfer=yes")
            xvfb, client = self.run_client(display, "--file-transfer=yes")
            r = pollwait(client, CLIENT_TIMEOUT)
            assert r is None, f"client terminated unexpectedly with code {exit_str(r)}"
            #send a file to this client:
            send_file_command = ["control", display, "send-file", f.name, "0", "*"]
            env = os.environ.copy()
            import secrets
            env["XPRA_USER_UUID"] = secrets.token_urlsafe(16)
            send_file = self.run_xpra(send_file_command, env=env)
            assert pollwait(send_file, CLIENT_TIMEOUT)==0, "send-file command returncode is %s" % send_file.poll()
            time.sleep(1)
            #now verify the file can be found in the download directory
            filename = os.path.join(os.path.expanduser(get_download_dir()), os.path.basename(f.name))
            assert os.path.exists(filename), "cannot find %s" % filename
            readback = load_binary_file(filename)
            assert readback==data, "file data corrupted"
            os.unlink(filename)
        finally:
            if client:
                client.terminate()
            if xvfb:
                xvfb.terminate()
            if server:
                server.terminate()
            f.close()

    def test_control_send_file(self):
        data = b"".join(uuid.uuid4().bytes for _ in range(100))
        self.do_test_control_send_file(data)


def main():
    if POSIX and not OSX:
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import sleep
from textwrap import wrap
from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.util.env import envint
from xpra.util.str_fn import csv
from xpra.util.thread import start_thread
from xpra.gtk.configure.common import run_gui
from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.gtk.widget import label
from xpra.platform.paths import get_image
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("gstreamer", "util")

STEP_DELAY = envint("XPRA_CONFIGURE_STEP_DELAY", 100)


def _set_labels_text(widgets, *messages: str) -> None:
    for i, widget in enumerate(widgets):
        if i < len(messages):
            widget.set_text(messages[i])
        else:
            widget.set_text("")


class ConfigureGUI(BaseGUIWindow):

    def __init__(self, parent: Gtk.Window | None = None):
        self.warning_pixbuf = get_image("warning.png")
        size = (800, 554)
        if self.warning_pixbuf:
            size = self.warning_pixbuf.get_width() + 20, self.warning_pixbuf.get_height() + 20
        self.layout = None
        self.warning_labels: list[Gtk.Label] = []
        self.labels: list[Gtk.Label] = []
        self.buttons: list[Gtk.Button] = []
        self.elements: Sequence[str] = ()
        super().__init__(
            "Configure Xpra's GStreamer Codecs",
            "gstreamer.png",
            wm_class=("xpra-configure-gstreamer-gui", "Xpra Configure GStreamer GUI"),
            default_size=size,
            header_bar=(False, False),
            parent=parent,
        )
        self.set_resizable(False)

    def add_layout(self) -> None:
        if not self.layout:
            layout = Gtk.Layout()
            layout.set_margin_top(0)
            layout.set_margin_bottom(0)
            layout.set_margin_start(0)
            layout.set_margin_end(0)
            self.layout = layout
            self.vbox.add(layout)

    def populate(self) -> None:
        self.set_box_margin(0, 0, 0, 0)
        self.add_layout()
        if self.warning_pixbuf:
            image = Gtk.Image.new_from_pixbuf(self.warning_pixbuf)
            self.layout.put(image, 0, 0)
        for i in range(3):
            lbl = label("", font="Sans 22")
            self.warning_labels.append(lbl)
            self.layout.put(lbl, 86, 70 + i * 40)
        for i in range(11):
            lbl = label("", font="Sans 12")
            self.layout.put(lbl, 78, 180 + i * 24)
            self.labels.append(lbl)

        self.set_warning_labels(
            "This tool can cause your system to crash,",
            "it may even damage hardware in rare cases.",
            "            Use with caution.",
        )
        self.set_labels("", "", *wrap(DISCLAIMER))
        self.add_buttons(
            ("Understood", self.detect_elements),
            ("Get me out of here", self.dismiss)
        )

    def set_warning_labels(self, *messages) -> None:
        _set_labels_text(self.warning_labels, *messages)

    def set_labels(self, *messages) -> None:
        _set_labels_text(self.labels, *messages)

    def add_buttons(self, *buttons) -> list[Gtk.Button]:
        # remove existing buttons:
        for button in self.buttons:
            self.layout.remove(button)
        i = 0
        x = 400 - 100 * len(buttons)
        for text, callback in buttons:
            button = Gtk.Button.new_with_label(text)
            button.connect("clicked", callback)
            button.show()
            self.buttons.append(button)
            self.layout.put(button, x + 200 * i, 450)
            i += 1
        return self.buttons

    def detect_elements(self, *_args) -> None:
        self.set_warning_labels(
            "Probing the GStreamer elements available,",
            "please wait.",
        )
        self.set_labels()
        self.add_buttons(("Abort and exit", self.dismiss))
        messages: list[str] = []

        def update_messages() -> None:
            sleep(STEP_DELAY / 1000)
            GLib.idle_add(self.set_labels, *messages)

        def add_message(msg: str) -> None:
            messages.append(msg)
            update_messages()

        def update_message(msg: str) -> None:
            messages[-1] = msg
            update_messages()

        def probe_elements() -> None:
            add_message("Loading the GStreamer bindings")
            try:
                gst = gi_import("Gst")
                update_message("loaded the GStreamer bindings")
                add_message("initializing GStreamer")
                gst.init(None)
                update_message("initialized GStreamer")
            except Exception as e:
                log("Warning failed to import GStreamer", exc_info=True)
                update_message(f"Failed to load GStreamer: {e}")
                return
            try:
                add_message("locating plugins")
                from xpra.gstreamer.common import import_gst, get_all_plugin_names
                import_gst()
                self.elements = get_all_plugin_names()
                update_message(f"found {len(self.elements)} elements")
            except Exception as e:
                log("Warning failed to load GStreamer plugins", exc_info=True)
                update_message(f"Failed to load plugins: {e}")
                return
            if not self.elements:
                update_message("no elements found - cannot continue")
                return
            pset = set(self.elements)
            need = {"capsfilter", "videoconvert", "videoscale", "queue"}
            missing = tuple(need - pset)
            if missing:
                add_message("some essential plugins are missing: " + csv(missing))
                add_message("install them then you can run this tool again")
                return
            add_message("all the essential plugins have been found")
            want = {"x264enc", "vp8enc", "vp9enc", "webmmux"}
            found = tuple(want & pset)
            if not found:
                add_message("no default encoders found,")
                add_message("install at least one plugin from: " + csv(want))
                add_message("then you can run this tool again")
                return
            try:
                from xpra.platform.shadow_server import GSTREAMER_CAPTURE_ELEMENTS
            except ImportError:
                pass
            else:
                want |= set(GSTREAMER_CAPTURE_ELEMENTS)
            missing = tuple(want - pset)
            if missing:
                add_message("some useful extra plugins you may want to install: " + csv(missing))

            GLib.timeout_add(STEP_DELAY * 6, self.add_buttons,
                             ("configure shadow mode", self.configure_shadow),
                             ("configure encoding", self.configure_encoding),
                             ("configure decoding", self.configure_decoding),
                             )

        start_thread(probe_elements, "probe-elements", daemon=True)

    def configure_encoding(self, *_args) -> None:
        pass

    def configure_decoding(self, *_args) -> None:
        pass

    def configure_shadow(self, *_args) -> None:
        pass


def main(_args) -> int:
    return run_gui(ConfigureGUI)


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv[1:]))
DISCLAIMER = """
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES(INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT(INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
""".replace("\n", "")



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import shlex
import signal
import os.path
from time import monotonic
from subprocess import Popen
from typing import Any
from collections.abc import Callable

from xpra.platform.features import COMMAND_SIGNALS
from xpra.util.child_reaper import getChildReaper, ProcInfo
from xpra.common import noop
from xpra.os_util import OSX, WIN32, gi_import
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, Ellipsizer
from xpra.util.env import envint, restore_script_env
from xpra.net.common import PacketType
from xpra.util.thread import start_thread
from xpra.scripts.parsing import parse_env, get_subcommands
from xpra.server.util import source_env, write_pid
from xpra.server import ServerExitMode
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("exec")

TERMINATE_DELAY = envint("XPRA_TERMINATE_DELAY", 1000) / 1000.0


class ChildCommandServer(StubServerMixin):
    """
    Mixin for servers that start subcommands,
    ie "--start=xterm"
    """

    PREFIX = "command"

    def __init__(self):
        self.child_display: str = ""
        self.start_commands = []
        self.start_late_commands = []
        self.start_child_commands = []
        self.start_child_late_commands = []
        self.start_after_connect = []
        self.start_after_connect_done = False
        self.start_child_after_connect = []
        self.start_on_connect = []
        self.start_child_on_connect = []
        self.start_on_last_client_exit = []
        self.start_child_on_last_client_exit = []
        self.exit_with_children: bool = False
        self.children_count: int = 0
        self.start_after_connect_done: bool = False
        self.start_new_commands: bool = False
        self.source_env: dict[str, str] = {}
        self.start_env: dict[str, str] = {}
        self.exec_cwd: str = os.getcwd()
        self.exec_wrapper: list[str] = []
        self.terminate_children: bool = False
        self.children_started: list[ProcInfo] = []
        self.child_reaper = None
        self.reaper_exit: Callable = self.reaper_exit_check
        # does not belong here...
        if not hasattr(self, "_exit_mode"):
            self._exit_mode = None
        if not hasattr(self, "session_name"):
            self.session_name: str = ""
        self.menu_provider = None
        # wait for main loop to run
        # and ensure that we don't run `late_start()` more than once,
        # even if __init__ is called multiple times:
        if not getattr(self, "late_start_requested", False):
            self.late_start_requested: bool = True
            GLib.idle_add(self.late_start)

    def late_start(self) -> None:
        def do_late_start():
            # wait for all threaded init to complete
            self.wait_for_threaded_init()
            self.exec_start_late_commands()

        start_thread(do_late_start, "command-late-start", daemon=True)

    def init(self, opts) -> None:
        self.exit_with_children = opts.exit_with_children
        self.terminate_children = opts.terminate_children
        self.start_new_commands = opts.start_new_commands
        self.start_commands = opts.start
        self.start_late_commands = opts.start_late
        self.start_child_commands = opts.start_child
        self.start_child_late_commands = opts.start_child_late
        self.start_after_connect = opts.start_after_connect
        self.start_child_after_connect = opts.start_child_after_connect
        self.start_on_connect = opts.start_on_connect
        self.start_child_on_connect = opts.start_child_on_connect
        self.start_on_last_client_exit = opts.start_on_last_client_exit
        self.start_child_on_last_client_exit = opts.start_child_on_last_client_exit
        if opts.exec_wrapper:
            self.exec_wrapper = shlex.split(opts.exec_wrapper)
        self.child_reaper = getChildReaper()
        self.source_env = source_env(opts.source_start)
        self.start_env = parse_env(opts.start_env)
        if self.start_new_commands:
            # may already have been initialized by servercore:
            from xpra.server.menu_provider import get_menu_provider
            self.menu_provider = self.menu_provider or get_menu_provider()
            self.menu_provider.on_reload.append(self.send_updated_menu)

    def threaded_setup(self) -> None:
        self.exec_start_commands()

        def set_reaper_callback():
            self.child_reaper.set_quit_callback(self.reaper_exit)
            self.child_reaper.check()

        GLib.idle_add(set_reaper_callback)

    def cleanup(self) -> None:
        if self.terminate_children and self._exit_mode not in (ServerExitMode.UPGRADE, ServerExitMode.EXIT):
            self.terminate_children_processes()
        # during cleanup, just ignore the reaper exit callback:
        self.reaper_exit = noop
        if self.menu_provider:
            self.menu_provider.cleanup()

    def get_server_features(self, _source) -> dict[str, Any]:
        return {
            "start-new-commands": self.start_new_commands,
            "exit-with-children": self.exit_with_children,
            "server-commands-signals": COMMAND_SIGNALS,
            "server-commands-info": not WIN32 and not OSX,
        }

    def _get_xdg_menu_data(self) -> dict[str, Any] | None:
        if not self.start_new_commands:
            return None
        assert self.menu_provider
        return self.menu_provider.get_menu_data()

    def get_caps(self, source) -> dict[str, Any]:
        caps: dict[str, Any] = {}
        if not source:
            return caps
        # don't assume we have a real ClientConnection object:
        wants = getattr(source, "wants", [])
        if "features" in wants and getattr(source, "ui_client", False):
            caps["xdg-menu"] = {}
            caps["subcommands"] = get_subcommands()
        return caps

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        self.exec_on_connect_commands()

    def send_initial_data(self, ss, caps: typedict, send_ui: bool, share_count: int) -> None:
        xdg_menu = getattr(ss, "xdg_menu", False)
        log(f"send_initial_data(..) {xdg_menu=}")
        if not xdg_menu:
            return
        # this method may block if the menus are still being loaded,
        # so do it in a throw-away thread:
        start_thread(self.send_xdg_menu_data, "send-xdg-menu-data", True, (ss,))

    def send_xdg_menu_data(self, ss) -> None:
        if ss.is_closed():
            return
        xdg_menu = self._get_xdg_menu_data() or {}
        self.do_send_xdg_menu_data(ss, xdg_menu)

    def do_send_xdg_menu_data(self, ss, xdg_menu):
        if ss.is_closed():
            return
        if not getattr(ss, "send_setting_change", False):
            return
        if not getattr(ss, "xdg_menu", False):
            return
        ss.send_setting_change("xdg-menu", xdg_menu)
        log(f"{len(xdg_menu)} menu data entries sent to {ss}")

    def send_updated_menu(self, xdg_menu) -> None:
        log("send_updated_menu(%s)", Ellipsizer(xdg_menu))
        for source in tuple(self._server_sources.values()):
            self.do_send_xdg_menu_data(source, xdg_menu)

    def get_info(self, _proto) -> dict[str, Any]:
        info: dict[Any, Any] = {
            "start": self.start_commands,
            "start-late": self.start_late_commands,
            "start-child": self.start_child_commands,
            "start-child-late": self.start_child_late_commands,
            "start-after-connect": self.start_after_connect,
            "start-child-after-connect": self.start_child_after_connect,
            "start-on-connect": self.start_on_connect,
            "start-child-on-connect": self.start_child_on_connect,
            "exit-with-children": self.exit_with_children,
            "start-after-connect-done": self.start_after_connect_done,
            "start-new": self.start_new_commands,
            "source-env": self.source_env,
            "start-env": self.start_env,
        }
        mp = self.menu_provider
        if mp:
            info.update(
                {
                    "start-menu": mp.get_menu_data(remove_icons=True, wait=False) or {},
                    "start-desktop-menu": mp.get_desktop_sessions(remove_icons=True) or {},
                }
            )
        for i, procinfo in enumerate(self.children_started):
            info[i] = procinfo.get_info()
        cinfo: dict[str, Any] = {ChildCommandServer.PREFIX: info}
        return cinfo

    def last_client_exited(self) -> None:
        self._exec_commands(self.start_on_last_client_exit, self.start_child_on_last_client_exit)

    def get_child_env(self) -> dict[str, str]:
        # subclasses may add more
        env = restore_script_env(super().get_child_env())
        env.update(self.source_env)
        env.update(self.start_env)
        if self.child_display:
            env["DISPLAY"] = self.child_display
        return env

    def get_full_child_command(self, cmd, use_wrapper=True) -> list[str]:
        # make sure we have it as a list:
        cmd = super().get_full_child_command(cmd, use_wrapper)
        if not use_wrapper or not self.exec_wrapper:
            return cmd
        return self.exec_wrapper + cmd

    def exec_start_late_commands(self) -> None:
        log("exec_start_late_commands() start-late=%s, start_child=%s",
            self.start_late_commands, self.start_child_late_commands)
        self._exec_commands(self.start_late_commands, self.start_child_late_commands)

    def exec_start_commands(self) -> None:
        log("exec_start_commands() start=%s, start_child=%s", self.start_commands, self.start_child_commands)
        self._exec_commands(self.start_commands, self.start_child_commands)

    def exec_after_connect_commands(self) -> None:
        log("exec_after_connect_commands() start=%s, start_child=%s",
            self.start_after_connect, self.start_child_after_connect)
        self._exec_commands(self.start_after_connect, self.start_child_after_connect)

    def exec_on_connect_commands(self) -> None:
        log("exec_on_connect_commands() start_after_connect_done=%s", self.start_after_connect_done)
        if not self.start_after_connect_done:  # pylint: disable=access-member-before-definition
            self.start_after_connect_done = True
            self.exec_after_connect_commands()
        log("exec_on_connect_commands() start=%s, start_child=%s", self.start_on_connect, self.start_child_on_connect)
        self._exec_commands(self.start_on_connect, self.start_child_on_connect)

    def _exec_commands(self, start_list, start_child_list) -> None:
        started = []
        if start_list:
            for x in start_list:
                if x:
                    proc = self.start_command(x, x, ignore=True)
                    if proc:
                        started.append(proc)
        if start_child_list:
            for x in start_child_list:
                if x:
                    proc = self.start_command(x, x, ignore=False)
                    if proc:
                        started.append(proc)
        procs = tuple(x for x in started if x is not None)
        if not self.session_name:
            GLib.idle_add(self.guess_session_name, procs)

    def start_command(self, name: str, child_cmd, ignore: bool = False, callback: Callable | None = None,
                      use_wrapper: bool = True, shell: bool = False, **kwargs):
        env = self.get_child_env()
        log("start_command%s exec_wrapper=%s, exec_cwd=%s",
            (name, child_cmd, ignore, callback, use_wrapper, shell, kwargs), self.exec_wrapper, self.exec_cwd)
        real_cmd = []
        cmd_str = None
        try:
            real_cmd = self.get_full_child_command(child_cmd, use_wrapper)
            log("full child command(%s, %s)=%s", child_cmd, use_wrapper, real_cmd)
            cmd_str = " ".join(real_cmd)
            # pylint: disable=consider-using-with
            proc = Popen(real_cmd, env=env, shell=shell, cwd=self.exec_cwd, **kwargs)
            procinfo = self.add_process(proc, name, real_cmd, ignore=ignore, callback=callback)
            is_ibus_daemon = real_cmd[0] == "ibus-daemon"
            cmd_info = "ibus-daemon" if is_ibus_daemon else cmd_str
            log.info(f"started command `{cmd_info}` with pid {proc.pid}")
            if not ignore:
                self.children_count += 1
            self.children_started.append(procinfo)
            session_dir = os.environ.get("XPRA_SESSION_DIR")
            if session_dir and not procinfo.dead:
                pidname = name.split(" ")[0]
                if pidname.find(os.sep) >= 0:
                    pidname = pidname.split(os.sep)[-1]
                pidfile = os.path.join(session_dir, f"{pidname}.pid")
                if not os.path.exists(pidfile):
                    procinfo.pidfile = pidfile
                    procinfo.pidinode = write_pid(pidfile, procinfo.pid)
                log(f"pidfile({name})={pidfile}, inode={procinfo.pidinode}, pid={procinfo.pid}")
            return proc
        except (OSError, ValueError) as e:
            log("start_command%s", (name, child_cmd, ignore, callback, use_wrapper, shell, kwargs), exc_info=True)
            log.error(f"Error spawning child {child_cmd!r}")
            if len(real_cmd) > 1:
                log.error(" using command:")
                log.error(f" `{cmd_str}`")
            log.error(f" {e}")
            return None

    def add_process(self, process, name: str, command, ignore: bool = False,
                    callback: Callable | None = None) -> ProcInfo:
        return self.child_reaper.add_process(process, name, command, ignore, callback=callback)

    @staticmethod
    def is_child_alive(proc) -> bool:
        return proc is not None and proc.poll() is None

    def reaper_exit_check(self) -> None:
        log(f"reaper_exit_check() exit_with_children={self.exit_with_children}")
        if self.exit_with_children and self.children_count:
            log.info("all children have exited and --exit-with-children was specified, exiting")
            GLib.idle_add(self.clean_quit)

    def terminate_children_processes(self) -> None:
        cl = tuple(self.children_started)
        self.children_started = []
        log(f"terminate_children_processes() children={cl}")
        if not cl:
            return
        wait_for = []
        self.child_reaper.poll()
        for procinfo in cl:
            proc = procinfo.process
            name = procinfo.name
            if self.is_child_alive(proc):
                wait_for.append(procinfo)
                log(f"child command {name!r} is still alive, calling terminate on {proc}")
                try:
                    proc.terminate()
                except Exception as e:
                    log(f"failed to terminate {proc}: {e}")
                    del e
        if not wait_for:
            return
        log(f"waiting for child commands to exit: {wait_for}")
        start = monotonic()
        while monotonic() - start < TERMINATE_DELAY and wait_for:
            self.child_reaper.poll()
            # this is called from the UI thread, we cannot sleep
            # sleep(1)
            wait_for = [procinfo for procinfo in wait_for if self.is_child_alive(procinfo.process)]
            log(f"still not terminated: {wait_for}")
        log("done waiting for child commands")

    def guess_session_name(self, procs=None) -> None:
        if not procs:
            return
        from xpra.scripts.server import IBUS_DAEMON_COMMAND  # pylint: disable=import-outside-toplevel
        if IBUS_DAEMON_COMMAND:
            ibus_daemon_cmd = shlex.split(IBUS_DAEMON_COMMAND)[0] or "ibus-daemon"
        else:
            ibus_daemon_cmd = ""
        # use the commands to define the session name:
        self.child_reaper.poll()
        cmd_names = []
        for proc in procs:
            proc_info = self.child_reaper.get_proc_info(proc.pid)
            if not proc_info:
                continue
            cmd = proc_info.command
            if self.exec_wrapper:
                # strip exec wrapper
                l = len(self.exec_wrapper)
                if len(cmd) > l and cmd[:l] == self.exec_wrapper:
                    cmd = cmd[l:]
            elif len(cmd) > 1 and cmd[0] in ("vglrun", "nohup",):
                cmd.pop(0)
            if ibus_daemon_cmd and cmd[0] == ibus_daemon_cmd:
                continue
            bcmd = os.path.basename(cmd[0])
            if bcmd not in cmd_names:
                cmd_names.append(bcmd)
        log(f"guess_session_name() commands={cmd_names}")
        if cmd_names:
            new_name = csv(cmd_names)
            if self.session_name != new_name:
                self.session_name = new_name
                self.mdns_update()

    def _process_start_command(self, proto, packet: PacketType) -> None:
        self._process_command_start(proto, packet)

    def _process_command_start(self, proto, packet: PacketType) -> None:
        log(f"start new command: {packet}")
        if not self.start_new_commands:
            log.warn("Warning: received start-command request,")
            log.warn(" but the feature is currently disabled")
            return
        name = str(packet[1])
        command = packet[2]
        ignore = bool(packet[3])
        cmd: str | tuple
        if isinstance(command, (list, tuple)):
            cmd = tuple(command)
        else:
            cmd = str(command)
        proc = self.start_command(name, cmd, ignore)
        if len(packet) >= 5:
            shared = packet[4]
            if proc and not shared:
                ss = self.get_server_source(proto)
                assert ss
                log(f"adding filter: pid={proc.pid} for {proto}")
                ss.add_window_filter("window", "pid", "=", proc.pid)
        log(f"process_start_command: proc={proc}")

    def _process_command_signal(self, _proto, packet: PacketType) -> None:
        pid = packet[1]
        signame = packet[2]
        if signame not in COMMAND_SIGNALS:
            log.warn("Warning: invalid signal received: '%s'", signame)
            return
        procinfo = self.child_reaper.get_proc_info(pid)
        if not procinfo:
            log.warn("Warning: command not found for pid %i", pid)
            return
        if procinfo.returncode is not None:
            log.warn("Warning: command for pid %i has already terminated", pid)
            return
        sigval = getattr(signal, signame, None)
        if not sigval:
            log.error(f"Error: signal {signame!r} not found!")
            return
        log.info(f"sending signal {signame!r} to pid {pid}")
        try:
            os.kill(pid, sigval)
        except Exception as e:
            log.error(f"Error sending signal {signame!r} to pid {pid}")
            log.estr(e)

    def init_packet_handlers(self) -> None:
        log("init_packet_handlers() COMMANDS_SIGNALS=%s, start new commands=%s",
            COMMAND_SIGNALS, self.start_new_commands)
        if COMMAND_SIGNALS:
            self.add_packets("command-signal")
        if self.start_new_commands:
            self.add_packets(f"{ChildCommandServer.PREFIX}-start")
            self.add_legacy_alias("start-command", f"{ChildCommandServer.PREFIX}-start")



#!/usr/bin/env python3

import gi

gi.require_version("Gtk", "3.0")  # @UndefinedVariable
gi.require_version('GdkX11', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GdkX11, GLib    # pylint: disable=wrong-import-position @UnresolvedImport
assert GdkX11    #this import has side-effects

width = 400
height = 200
class MoveWindow(Gtk.Window):
    def __init__(self, window_type):
        super().__init__(type=window_type)
        self.set_size_request(width, height)
        self.connect("delete_event", Gtk.main_quit)
        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        hbox = Gtk.HBox(homogeneous=False, spacing=0)
        vbox.pack_start(hbox, expand=False, fill=False, padding=10)

        gtk_btn = Gtk.Button(label="move+resize via GTK")
        hbox.pack_start(gtk_btn, expand=False, fill=False, padding=10)
        gtk_btn.connect('clicked', self.moveresize_GTK)

        x11_btn = Gtk.Button(label="move+resize via x11")
        hbox.pack_start(x11_btn, expand=False, fill=False, padding=10)
        x11_btn.connect('clicked', self.moveresize_X11)

        self.add(vbox)
        self.show_all()
        GLib.timeout_add(5*1000, self.moveresize_GTK)
        GLib.timeout_add(10*1000, self.moveresize_X11)

    def moveresize_GTK(self, *_args):
        new_x, new_y, new_width, new_height = self.get_new_geometry()
        self.move(new_x, new_y)
        self.resize(new_width, new_height)

    def moveresize_X11(self, *_args):
        new_x, new_y, new_width, new_height = self.get_new_geometry()
        from xpra.x11.gtk.display_source import init_gdk_display_source
        init_gdk_display_source()
        from xpra.x11.bindings.window import constants, X11WindowBindings  #@UnresolvedImport
        X11Window = X11WindowBindings()
        root = self.get_window().get_screen().get_root_window()
        root_xid = root.get_xid()
        xwin = self.get_window().get_xid()
        SubstructureNotifyMask = constants["SubstructureNotifyMask"]
        SubstructureRedirectMask = constants["SubstructureRedirectMask"]
        event_mask = SubstructureNotifyMask | SubstructureRedirectMask
        X11Window.sendClientMessage(root_xid, xwin, False, event_mask, "_NET_MOVERESIZE_WINDOW",
                                    1+2**8+2**9+2**10+2**11, new_x, new_y, new_width, new_height)

    def get_new_geometry(self):
        x, y = self.get_position()
        width, height = self.get_size()
        from xpra.gtk.util import get_default_root_window
        maxx, maxy = get_default_root_window().get_geometry()[2:4]
        new_x = (x+100) % (maxx-width)
        new_y = (y+100) % (maxy-height)
        width, height = self.get_size()
        return new_x, new_y, (width + 100) % (maxx-x), (height + 100) % (maxy-y)


def main():
    MoveWindow(Gtk.WindowType.TOPLEVEL)
    MoveWindow(Gtk.WindowType.POPUP)
    Gtk.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import dbus.service

from xpra.server.dbus.server import DBUS_Server, INTERFACE
from xpra.log import Logger

log = Logger("dbus", "server")


class X11_DBUS_Server(DBUS_Server):

    @dbus.service.method(INTERFACE)
    def SyncXvfb(self):
        self.server.do_repaint_root_overlay()

    @dbus.service.method(INTERFACE)
    def ResetXSettings(self):
        self.server.update_all_server_settings(True)

    @dbus.service.method(INTERFACE, in_signature='ii')
    def SetDPI(self, xdpi, ydpi):
        self.server.set_dpi(xdpi, ydpi)

    @dbus.service.method(INTERFACE, in_signature='ii', out_signature='ii')
    def SetScreenSize(self, width, height):
        return self.server.set_screen_size(width, height)

    @dbus.service.method(INTERFACE)
    def ShowAllWindows(self):
        self.server.show_all_windows()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.common import noop
from xpra.util.objects import AdHocStruct
from xpra.client.mixins.tray import TrayClient
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class AudioClientTest(ClientMixinTest):

	def test_tray(self):
		def _TrayClient():
			x = TrayClient()
			x.after_handshake = noop
			def get_tray_menu_helper_class():
				return None
			x.get_tray_menu_helper_class = get_tray_menu_helper_class
			return x
		opts = AdHocStruct()
		opts.tray = True
		opts.delay_tray = 0
		opts.tray_icon = ""
		self._test_mixin_class(_TrayClient, opts)

def main():
	unittest.main()


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import unittest

from xpra.util.env import OSEnvContext
from xpra.net import common
from xpra.log import set_global_logging_handler


def mlp(sending, packet_type, *args) -> None:
    packet = [packet_type] + list(args)
    common.may_log_packet(sending, packet_type, packet)


class TestLogPackets(unittest.TestCase):

    def info(self, *args):
        self.log_messages.append(args)

    def setup_log_intercept(self):
        # get_log_packets, may_log_packet, log
        self.log_messages = []
        set_global_logging_handler(self.info)

    def lm(self, n=0):
        assert len(self.log_messages) == n, "expected %i log messages, got %i" % (n, len(self.log_messages))
        self.log_messages = []

    def test_env_log(self):
        with OSEnvContext():
            os.environ["XPRA_LOG_PACKETS"] = "info,ping,-bell"
            for log_packet_type in (False, True):
                os.environ["XPRA_LOG_PACKET_TYPE"] = str(int(log_packet_type))
                common.init()
                inc = int(log_packet_type)
                logged = 1+inc
                for sending in (True, False):
                    self.setup_log_intercept()

                    def t(packet_type, *args) -> None:
                        mlp(sending, packet_type, *args)   # pylint: disable=cell-var-from-loop

                    self.lm()
                    t("hello", {})
                    self.lm(inc)
                    t("info", {"foo": "bar"})
                    self.lm(logged)
                    t("ping", 1, 2, 3)
                    self.lm(logged)
                    t("ping-echo", 1, 2, 3)
                    self.lm(inc)
                    t("bell", 100)
                    self.lm(inc)
                    t("info", "0"*common.PACKET_LOG_MAX_SIZE*2)
                    assert len(self.log_messages[-1]) <= common.PACKET_LOG_MAX_SIZE
                    self.lm(logged)

    def test_default_nolog(self):
        with OSEnvContext():
            os.environ.pop("XPRA_LOG_PACKETS", None)
            self.setup_log_intercept()
            for pt in common.PACKET_TYPES:
                mlp(True, pt, (1, 2))
                mlp(False, pt, (1, 2))
                self.lm(0)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os.path
from types import ModuleType
from typing import Any
from importlib import import_module
from collections.abc import Sequence, Iterable

from xpra.common import noop
from xpra.util.str_fn import csv, print_nested_dict, pver
from xpra.util.env import envbool, numpy_import_context
from xpra.os_util import OSX, WIN32
from xpra.util.version import parse_version
from xpra.codecs.constants import HELP_ORDER
from xpra.log import Logger, enable_color, LOG_FORMAT, NOPREFIX_FORMAT
log = Logger("codec", "loader")


# these codecs may well not load because we
# do not require the libraries to be installed
NOWARN = [
    "nvenc", "nvdec", "enc_nvjpeg",
    "dec_nvjpeg", "nvfbc", "dec_openh264",
    "enc_gstreamer", "dec_gstreamer",
    "csc_cython", "dec_avif", "enc_avif",
    "enc_amf",
]

SELFTEST = envbool("XPRA_CODEC_SELFTEST", True)
FULL_SELFTEST = envbool("XPRA_CODEC_FULL_SELFTEST", False)

CODEC_FAIL_IMPORT = os.environ.get("XPRA_CODEC_FAIL_IMPORT", "").split(",")
CODEC_FAIL_SELFTEST = os.environ.get("XPRA_CODEC_FAIL_SELFTEST", "").split(",")

log(f"codec loader settings: {SELFTEST=}, {FULL_SELFTEST=}, {CODEC_FAIL_IMPORT=}, {CODEC_FAIL_SELFTEST=}")


SKIP_LIST: Sequence[str] = ()
if OSX:
    SKIP_LIST = ("avif", "nvenc", "nvdec", "nvjpeg")


def autoprefix(prefix: str, name: str) -> str:
    return (name if (name.startswith(prefix) or name.endswith(prefix)) else f"{prefix}_{name}").replace("-", "_")


def filt(*values) -> tuple[str, ...]:
    return tuple(x for x in values if all(x.find(s) < 0 for s in SKIP_LIST))


def gfilt(generator) -> tuple[str, ...]:
    return filt(*generator)


CSC_CODECS: Sequence[str] = gfilt(f"csc_{x}" for x in ("cython", "libyuv"))
ENCODER_CODECS: Sequence[str] = gfilt(f"enc_{x}" for x in (
    "rgb", "pillow", "spng", "webp", "jpeg", "nvjpeg", "avif",
))
ENCODER_VIDEO_CODECS: Sequence[str] = gfilt(autoprefix("enc", x) for x in (
    "vpx", "x264", "openh264", "nvenc", "gstreamer", "amf", "remote",
))
DECODER_CODECS: Sequence[str] = gfilt(f"dec_{x}" for x in (
    "pillow", "spng", "webp", "jpeg", "nvjpeg", "avif", "gstreamer",
))
DECODER_VIDEO_CODECS: Sequence[str] = gfilt(autoprefix("dec", x) for x in (
    "vpx", "openh264", "nvdec",
))
SOURCES: Sequence[str] = filt("v4l2", "evdi", "drm", "nvfbc")

ALL_CODECS: Sequence[str] = filt(*set(
    CSC_CODECS + ENCODER_CODECS + ENCODER_VIDEO_CODECS + DECODER_CODECS + DECODER_VIDEO_CODECS + SOURCES)
)


codec_errors: dict[str, str] = {}
codecs: dict[str, ModuleType] = {}


def should_warn(name: str) -> bool:
    if name in NOWARN:
        return False
    if name in ("csc_cython", "dec_avif") or name.find("enc") >= 0 or name.startswith("nv"):
        try:
            import_module("xpra.server")
        except ImportError:
            # no server support,
            # probably a 'Light' build
            return False
    return True


def pillow_import_block() -> None:
    # the only ones we want to keep:
    # "Bmp", "Gif", "Ppm", "Png", "Jpeg", "Xpm"
    for image_plugin in (
        "Blp", "Cur", "Pcx", "Dcx", "Dds", "Eps", "Fits", "Fli",
        "Fpx", "Ftex", "Gbr", "Jpeg2K", "Icns", "Ico",
        "Im", "Imt",
        "Iptc", "McIdas", "Mic", "Mpeg", "Tiff", "Mpo", "Msp",
        "Palm", "Pcd", "Pdf", "Pixar", "Psd", "Qoi",
        "Sgi", "Spider", "Sun", "Tga",
        "Wmf", "Xbm", "XVThumb",
    ):
        sys.modules[f"PIL.{image_plugin}ImagePlugin"] = None


PIL_BLOCK = envbool("XPRA_PIL_BLOCK", True)
if PIL_BLOCK:
    pillow_import_block()


def codec_import_check(name: str, description: str, top_module: str, class_module: str,
                       attrs: Sequence[str], options: dict):
    log(f"{name}:")
    log(" codec_import_check%s", (name, description, top_module, class_module, attrs))
    if any(name.find(s) >= 0 for s in SKIP_LIST):
        log(f" skipped from list: {csv(SKIP_LIST)}")
        return None
    try:
        if name in CODEC_FAIL_IMPORT:
            raise ImportError("codec found in fail import list")
        module = import_module(top_module)
        log(f"imported {module!r}")
    except ImportError as e:
        log(f"failed to import {name} ({description})")
        log("", exc_info=True)
        codec_errors[name] = str(e)
        return None
    except Exception as e:
        log.warn(f" cannot load {name} ({description}):", exc_info=True)
        codec_errors[name] = str(e)
        return None
    attr = None
    try:
        log(f" {top_module} found, will check for {attrs} in {class_module}")
        ic = import_module(class_module)

        init_module = getattr(ic, "init_module", noop)
        log(f"{ic}.init_module={init_module}")
        init_module(options)

        if log.is_debug_enabled():
            # try to enable debugging on the codec's own logger:
            module_logger = getattr(ic, "log", None)
            log(f"{class_module}.log={module_logger}")
            if module_logger:
                module_logger.enable_debug()

        for attr in attrs:
            try:
                clazz = getattr(ic, attr)
            except AttributeError:
                raise ImportError(f"cannot find {attr!r} in {ic}") from None
            log(f"{class_module}.{attr}={clazz}")

        selftest = getattr(ic, "selftest", None)
        log(f"{name}.selftest={selftest}")
        if SELFTEST and selftest:
            if name in CODEC_FAIL_SELFTEST:
                raise ImportError("codec found in fail selftest list")
            try:
                selftest(FULL_SELFTEST)
            except Exception as e:
                log(f"{selftest} failed", exc_info=True)
                if not isinstance(e, ImportError):
                    log.warn(f"Warning: {name} failed its self test")
                    for x in str(e).splitlines():
                        log.warn(f" {x}")
                return None

        log(f" found {name} : {ic}")
        codecs[name] = ic
        return ic
    except ImportError as e:
        codec_errors[name] = str(e)
        log_fn = log.error if should_warn(name) else log.debug
        log_fn(f"Error importing {name} ({description})")
        log_fn(f" {e}")
        log("", exc_info=True)
    except Exception as e:
        codec_errors[name] = str(e)
        if attr:
            log.warn(f" cannot load {name} ({description}): {attr} missing from {class_module}", exc_info=True)
        else:
            log.warn(f" cannot load {name} ({description})", exc_info=True)
    return None


codec_versions: dict[str, Iterable[Any]] = {}


def add_codec_version(name: str, top_module, version: str = "get_version()", alt_version: str = "__version__") -> None:
    try:
        fieldnames = [x for x in (version, alt_version) if x is not None]
        for fieldname in fieldnames:
            f = fieldname
            if f.endswith("()"):
                f = version[:-2]
            module = import_module(top_module)
            if not hasattr(module, f):
                continue
            v = getattr(module, f)
            log(f"{module}.{f}={v}")
            if fieldname.endswith("()") and v:
                log(f"calling {v}")
                v = v()
            codec_versions[name] = parse_version(v)
            # optional info:
            if hasattr(module, "get_info"):
                info = getattr(module, "get_info")
                log(f" {name} {top_module}.{info}={info()}")
            return
        if name in codecs:
            log.warn(f" cannot find %s in {top_module}", " or ".join(fieldnames))
        else:
            log(f" no version information for missing codec {name}")
    except ImportError as e:
        # not present
        log(f" cannot import {name}: {e}")
        log("", exc_info=True)
    except Exception as e:
        log.warn(f"error during {name} codec import: %s", e)
        log.warn("", exc_info=True)


platformname = sys.platform.rstrip("0123456789")


CODEC_OPTIONS: dict[str, tuple[str, str, str, str]] = {
    # encoders:
    "enc_rgb"       : ("RGB encoder",       "argb",         "encoder", "encode"),
    "enc_pillow"    : ("Pillow encoder",    "pillow",       "encoder", "encode"),
    "enc_spng"      : ("png encoder",       "spng",         "encoder", "encode"),
    "enc_webp"      : ("webp encoder",      "webp",         "encoder", "encode"),
    "enc_jpeg"      : ("JPEG encoder",      "jpeg",         "encoder", "encode"),
    "enc_avif"      : ("avif encoder",      "avif",         "encoder", "encode"),
    "enc_nvjpeg"    : ("nvjpeg encoder",    "nvidia.nvjpeg", "encoder", "encode"),
    # video encoders:
    "enc_vpx"       : ("vpx encoder",       "vpx",          "encoder", "Encoder"),
    "enc_x264"      : ("x264 encoder",      "x264",         "encoder", "Encoder"),
    "enc_openh264"  : ("openh264 encoder",  "openh264",     "encoder", "Encoder"),
    "nvenc"         : ("nvenc encoder",     "nvidia.nvenc", "encoder", "Encoder"),
    "enc_gstreamer" : ("gstreamer encoder", "gstreamer",    "encoder", "Encoder"),
    "enc_amf"       : ("amf encoder",       "amf",          "encoder", "Encoder"),
    "enc_remote"    : ("remote encoder",    "remote",       "encoder", "Encoder"),
    # csc:
    "csc_libyuv"    : ("libyuv colorspace conversion", "libyuv", "converter", "Converter"),
    "csc_cython"    : ("cython colorspace conversion", "csc_cython", "converter", "Converter"),
    # decoders:
    "dec_pillow"    : ("Pillow decoder",    "pillow",       "decoder", "decompress"),
    "dec_spng"      : ("png decoder",       "spng",         "decoder", "decompress"),
    "dec_webp"      : ("webp decoder",      "webp",         "decoder", "decompress_to_rgb,decompress_to_yuv"),
    "dec_jpeg"      : ("JPEG decoder",      "jpeg",         "decoder", "decompress_to_rgb,decompress_to_yuv"),
    "dec_avif"      : ("avif decoder",      "avif",         "decoder", "decompress"),
    "dec_nvjpeg"    : ("nvjpeg decoder",    "nvidia.nvjpeg", "decoder", "decompress"),
    # video decoders:
    "dec_vpx"       : ("vpx decoder",       "vpx",          "decoder", "Decoder"),
    "dec_openh264"  : ("openh264 decoder",  "openh264",     "decoder", "Decoder"),
    "nvdec"         : ("nvdec decoder",     "nvidia.nvdec", "decoder", "Decoder"),
    "dec_gstreamer" : ("gstreamer decoder", "gstreamer",    "decoder", "Decoder"),
    # sources:
    "v4l2"          : ("v4l2 source",       "v4l2",         "virtual", "VirtualWebcam"),
    "evdi"          : ("evdi source",       "evdi",         "capture", "EvdiDevice"),
    "drm"           : ("drm device query",  "drm",          "drm",      "query"),
    "nvfbc"         : ("NVIDIA Capture SDK", "nvidia.nvfbc", f"capture_{platformname}", "NvFBC_SysCapture"),
}

NOLOAD: list[str] = []
if OSX:
    # none of the nvidia codecs are available on MacOS,
    # so don't bother trying:
    NOLOAD += ["nvenc", "enc_nvjpeg", "dec_nvjpeg", "nvfbc"]
if OSX or WIN32:
    # these sources can only be used on Linux
    # (and maybe on some BSDs?)
    NOLOAD += ["v4l2", "evdi", "drm"]


def load_codec(name: str, options: dict | None = None):
    log("load_codec(%s, %s)", name, options)
    name = name.replace("-", "_")
    if not has_codec(name):
        try:
            description, top_module, class_module, attrs_names = CODEC_OPTIONS[name]
            attrs = attrs_names.split(",")
        except KeyError:
            log("load_codec(%s)", name, exc_info=True)
            log.error("Error: invalid codec name '%s'", name)
            return None
        xpra_top_module = f"xpra.codecs.{top_module}"
        xpra_class_module = f"{xpra_top_module}.{class_module}"
        if codec_import_check(name, description, xpra_top_module, xpra_class_module, attrs, options or {}):
            version_name = name
            if name.startswith("enc_") or name.startswith("dec_") or name.startswith("csc_"):
                version_name = name[4:]
            add_codec_version(version_name, xpra_class_module)
    return get_codec(name)


def load_codecs(encoders=True, decoders=True, csc=True, video=True, sources=False) -> Sequence[str]:
    log("loading codecs")
    loaded: list[str] = []

    def load(*names):
        for name in names:
            if has_codec(name):
                continue
            if name in NOLOAD:
                log(f"{name} is in the NOLOAD list for this platform: {NOLOAD}")
                continue
            load_codec(name)
            if has_codec(name) and name not in loaded:
                loaded.append(name)

    if encoders:
        load(*ENCODER_CODECS)
        if video:
            load(*ENCODER_VIDEO_CODECS)
    if csc and video:
        load(*CSC_CODECS)
    if decoders:
        load(*DECODER_CODECS)
        if video:
            load(*DECODER_VIDEO_CODECS)
    if sources:
        load(*SOURCES)
    log("done loading codecs: %s", loaded)
    return tuple(loaded)


def unload_codecs() -> None:
    global codecs
    log(f"unload_codecs() {codecs=}")
    copy = codecs
    codecs = {}
    for name, module in copy.items():
        cleanup_module = getattr(module, "cleanup_module", noop)
        try:
            log(f"{name} cleanup_module={cleanup_module}")
            cleanup_module()
        except RuntimeError as e:
            log(f"{cleanup_module}()", exc_info=True)
            log.warn(f"Warning: error during {name!r} module cleanup")
            log.warn(f" {e}")


def show_codecs(show: Iterable[str] = ()) -> None:
    for name in sorted(show or ALL_CODECS):
        log(f"* {name.ljust(20)} : {str(name in codecs).ljust(10)} {codecs.get(name, '')}")
    log("codecs versions:")
    for name in (show or codec_versions.keys()):
        version = codec_versions.get(name, "")
        log(f"* {name.ljust(20)} : {version}")


def get_codec_error(name: str) -> str:
    return codec_errors.get(name, "")


def get_codec(name: str):
    if name not in CODEC_OPTIONS:
        log.warn(f"Warning: invalid codec name {name}")
    return codecs.get(name)


def get_codec_version(name: str):
    return codec_versions.get(name)


def has_codec(name: str) -> bool:
    return name in codecs


def get_rgb_compression_options() -> list[str]:
    # pylint: disable=import-outside-toplevel
    from xpra.net import compression
    compressors = compression.get_enabled_compressors()
    compressors = tuple(x for x in compressors if x != "brotli")
    RGB_COMP_OPTIONS: list[str] = ["Raw RGB"]
    if compressors:
        RGB_COMP_OPTIONS += ["/".join(compressors)]
    return RGB_COMP_OPTIONS


def get_encoding_name(encoding: str) -> str:
    ENCODINGS_TO_NAME : dict[str, str] = {
        "auto"    : "automatic",
        "stream"  : "video stream",
        "h264"    : "H.264",
        "h265"    : "H.265",
        "mpeg4"   : "MPEG4",
        "vp8"     : "VP8",
        "webp"    : "WebP",
        "vp9"     : "VP9",
        "png"     : "PNG (24/32bpp)",
        "png/P"   : "PNG (8bpp colour)",
        "png/L"   : "PNG (8bpp grayscale)",
        "jpeg"    : "JPEG",
        "jpega"   : "JPEG with alpha",
        "avif"    : "AVIF",
        "av1"     : "AV1",
        "rgb"     : " + ".join(get_rgb_compression_options()) + " (24/32bpp)",
    }
    return ENCODINGS_TO_NAME.get(encoding, encoding)


def get_encoding_help(encoding: str) -> str:
    # pylint: disable=import-outside-toplevel
    from xpra.net import compression
    compressors = [x for x in compression.get_enabled_compressors()
                   if x not in ("brotli", "none")]
    compressors_str = ""
    if compressors:
        compressors_str = ", may be compressed using "+(" or ".join(compressors))+" "
    return {
        "auto"    : "automatic mode (recommended)",
        "stream"  : "video stream",
        "grayscale" : "same as 'auto' but in grayscale mode",
        "h264"    : "H.264 video codec",
        "h265"    : "H.265 (HEVC) video codec (not recommended)",
        "vp8"     : "VP8 video codec",
        "vp9"     : "VP9 video codec",
        "mpeg4"   : "MPEG-4 video codec",
        "png"     : "Portable Network Graphics (lossless, 24bpp or 32bpp for transparency)",
        "png/P"   : "Portable Network Graphics (lossy, 8bpp colour)",
        "png/L"   : "Portable Network Graphics (lossy, 8bpp grayscale)",
        "webp"    : "WebP compression (supports lossless and lossy modes)",
        "jpeg"    : "JPEG lossy compression",
        "jpega"   : "JPEG lossy compression, with alpha channel",
        "avif"    : "AVIF: AV1 Image File Format",
        "av1"     : "AV1: AOMedia Video 1",
        "rgb"     : f"Raw RGB pixels, lossless {compressors_str}(24bpp or 32bpp for transparency)",
        "scroll"  : "motion vectors, supplemented with picture codecs",
    }.get(encoding, "")


def encodings_help(encodings: Iterable[str]) -> list[str]:
    h = []
    for e in HELP_ORDER:
        if e in encodings:
            h.append(encoding_help(e))
    return h


def encoding_help(encoding: str) -> str:
    ehelp = get_encoding_help(encoding) or ""
    return encoding.ljust(12) + ehelp


def main(args) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("Loader", "Encoding Info"):
        verbose = "-v" in args or "--verbose" in args
        args = [x for x in args if x not in ("-v", "--verbose")]
        format_string = NOPREFIX_FORMAT
        if verbose:
            global FULL_SELFTEST
            FULL_SELFTEST = True
            format_string = LOG_FORMAT
            log.enable_debug()
            from xpra.codecs.checks import log as check_log
            check_log.enable_debug()
        enable_color(format_string=format_string)

        from xpra.os_util import gi_import
        GLib = gi_import("GLib")
        main_loop = GLib.MainLoop()

        def load_in_thread() -> None:
            codecs = do_main_load(args)
            print_codecs(codecs)
            unload_codecs()
            main_loop.quit()

        from xpra.util.thread import start_thread
        start_thread(load_in_thread, "threaded loader")

        main_loop.run()
        return 0


def do_main_load(args) -> Sequence[str]:
    if len(args) > 1:
        names = []
        for x in args[1:]:
            name = x.lower().replace("-", "_")
            if name not in CODEC_OPTIONS:
                loose_matches = tuple(o for o in (
                    f"enc_{name}", f"dec_{name}", f"csc_{name}"
                ) if o in CODEC_OPTIONS)
                if len(loose_matches) == 1:
                    name = loose_matches[0]
                elif len(loose_matches) > 1:
                    log.warn(f"{x} matches: "+csv(loose_matches))
            load_codec(name)
            names.append(name)
        return tuple(names)
    try:
        load_codecs(sources=True)
    except KeyboardInterrupt:
        return ()
    # not really a codec, but gets used by codecs, so include version info:
    with numpy_import_context("codec loader"):
        add_codec_version("numpy", "numpy")
    return ALL_CODECS


def print_codecs(list_codecs: Sequence[str]) -> None:
    # use another logger for printing the results,
    # and use debug level by default, which shows up as green
    out = Logger("encoding")
    out.enable_debug()
    enable_color(format_string=NOPREFIX_FORMAT)
    out.info("modules found:")
    for name in sorted(list_codecs):
        mod : ModuleType | None = codecs.get(name)
        f = str(mod)
        if mod and hasattr(mod, "__file__"):
            f = getattr(mod, "__file__", f)
            if f.startswith(os.getcwd()):
                f = f[len(os.getcwd()):]
                if f.startswith(os.path.sep):
                    f = f[1:]
        if name in NOLOAD and not f:
            # don't show codecs from the NOLOAD list
            continue
        if mod:
            out.info(f"* {name.ljust(20)} : {f}")
            if log.is_debug_enabled():
                try:
                    if name.find("csc") >= 0:
                        cs = []
                        for spec in list(mod.get_specs()):
                            cs.append(spec.input_colorspace)
                        out(f"                         colorspaces: {csv(list(set(cs)))}")
                    elif name.find("enc") >= 0 or name.find("dec") >= 0:
                        encodings = mod.get_encodings()
                        out(f"                         encodings: {csv(encodings)}")
                    try:
                        i = mod.get_info()
                        for k, v in sorted(i.items()):
                            out(f"                         {k} = {v}")
                    except (AttributeError, RuntimeError):
                        pass
                except Exception as e:
                    log(f"{mod}", exc_info=True)
                    log.error(f"error getting extra information on {name}: {e}")
        elif name in codec_errors:
            write = out.error if should_warn(name) else out.debug
            write(f"* {name.ljust(20)} : {codec_errors[name]}")
    out("")
    out.info("codecs versions:")

    def forcever(v) -> str:
        return pver(v, numsep=".", strsep=".").lstrip("v")
    print_nested_dict(codec_versions, vformat=forcever, print_fn=out.info)


if __name__ == "__main__":
    main(sys.argv)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from random import randint
from ctypes import c_ubyte, c_uint32
from typing import Any

from xpra.common import roundup, noop, PaintCallback
from xpra.util.env import envbool, shellsub
from xpra.util.str_fn import csv
from xpra.util.objects import typedict
from xpra.os_util import get_group_id, get_int_uuid, WIN32, POSIX
from xpra.scripts.config import FALSE_OPTIONS
from xpra.util.stats import std_unit
from xpra.log import Logger

log = Logger("mmap")

ALWAYS_WRAP = envbool("XPRA_MMAP_ALWAYS_WRAP", False)
MMAP_GROUP = os.environ.get("XPRA_MMAP_GROUP", "xpra")
MADVISE = envbool("XPRA_MMAP_MADVISE", True)

DEFAULT_TOKEN_BYTES: int = 128

"""
Utility functions for communicating via mmap
"""


def get_socket_group(socket_filename) -> int:
    if isinstance(socket_filename, str) and os.path.exists(socket_filename):
        s = os.stat(socket_filename)
        return s.st_gid
    log.warn(f"Warning: missing valid socket filename {socket_filename!r} to set mmap group")
    return -1


def xpra_group() -> int:
    if POSIX:
        try:
            groups = os.getgroups()
            group_id = get_group_id(MMAP_GROUP)
            log("xpra_group() group(%s)=%s, groups=%s", MMAP_GROUP, group_id, groups)
            if group_id and group_id in groups:
                return group_id
        except Exception:
            log("xpra_group()", exc_info=True)
    return 0


def madvise(mmap_area) -> None:
    import mmap
    if not hasattr(mmap, "madvise"):
        log("no mmap.madvise() on this platform")
        return
    MADVISE_FLAGS = os.environ.get("XPRA_MMAP_MADVISE_FLAGS", "SEQUENTIAL,DONTFORK,UNMERGEABLE,DONTDUMP").split(",")
    log(f"setting MADVISE_FLAGS={MADVISE_FLAGS}")
    try:
        for flag in MADVISE_FLAGS:
            flag_value = getattr(mmap, f"MADV_{flag}", 0)
            log(f"MADV_{flag}={flag_value}")
            if flag_value:
                mmap_area.madvise(flag_value)
    except OSError as e:
        log(f"{mmap_area}.madvise(..)")
        log.error("Error: failed to set madvise flags %s", csv(MADVISE_FLAGS))
        log.estr(e)


def init_client_mmap(mmap_group=None, socket_filename: str = "", size: int = 128 * 1024 * 1024, filename: str = "") \
        -> tuple[bool, bool, Any, int, Any, str]:
    """
        Initializes a mmap area, writes the token in it and returns:
            (success flag, mmap_area, mmap_size, temp_file, mmap_filename)
        The caller must keep hold of temp_file to ensure it does not get deleted!
        This is used by the client.
    """

    def rerr() -> tuple[bool, bool, Any, int, Any, str]:
        return False, False, None, 0, None, ""

    log("init_client_mmap%s", (mmap_group, socket_filename, size, filename))
    mmap_filename = filename
    mmap_temp_file = None
    delete = True

    def validate_size(size: int) -> None:
        if size < 64 * 1024 * 1024:
            raise ValueError("mmap size is too small: %sB (minimum is 64MB)" % std_unit(size))
        if size > 16 * 1024 * 1024 * 1024:
            raise ValueError("mmap is too big: %sB (maximum is 4GB)" % std_unit(size))

    try:
        import mmap
        unit = max(4096, mmap.PAGESIZE)
        # add 8 bytes for the mmap area control header zone:
        mmap_size = roundup(size + 8, unit)
        if WIN32:
            validate_size(mmap_size)
            if not filename:
                from xpra.os_util import get_hex_uuid
                filename = "xpra-%s" % get_hex_uuid()
            mmap_filename = filename
            mmap_area = mmap.mmap(0, mmap_size, filename)
            # not a real file:
            delete = False
        else:
            assert POSIX
            if filename:
                if os.path.exists(filename):
                    fd = os.open(filename, os.O_EXCL | os.O_RDWR)
                    mmap_size = os.path.getsize(mmap_filename)
                    validate_size(mmap_size)
                    # mmap_size = 4*1024*1024    #size restriction needed with ivshmem
                    delete = False
                    log.info("Using existing mmap file '%s': %sMB", mmap_filename, mmap_size // 1024 // 1024)
                else:
                    validate_size(mmap_size)
                    flags = os.O_CREAT | os.O_EXCL | os.O_RDWR
                    try:
                        fd = os.open(filename, flags)
                        log("os.open(%r, %s)=%i", filename, flags, fd)
                        mmap_temp_file = None  # os.fdopen(fd, 'w')
                        mmap_filename = filename
                    except FileExistsError:
                        log.error("Error: the mmap file '%s' already exists", filename)
                        return rerr()
            else:
                validate_size(mmap_size)
                import tempfile
                from xpra.platform.paths import get_mmap_dir
                mmap_dir = get_mmap_dir()
                subs = os.environ.copy()
                subs |= {
                    "UID": str(os.getuid()),
                    "GID": str(os.getgid()),
                    "PID": str(os.getpid()),
                }
                mmap_dir = shellsub(mmap_dir, subs)
                if mmap_dir and not os.path.exists(mmap_dir):
                    os.mkdir(mmap_dir, 0o700)
                if not mmap_dir or not os.path.exists(mmap_dir):
                    raise RuntimeError("mmap directory %s does not exist!" % mmap_dir)
                # create the mmap file, the mkstemp that is called via NamedTemporaryFile ensures
                # that the file is readable and writable only by the creating user ID
                try:
                    temp = tempfile.NamedTemporaryFile(prefix="xpra.", suffix=".mmap", dir=mmap_dir)
                except OSError as e:
                    log.error("Error: cannot create mmap temporary file:")
                    log.estr(e)
                    return rerr()
                # keep a reference to it, so it does not disappear!
                mmap_temp_file = temp
                mmap_filename = temp.name
                fd = temp.file.fileno()
            # set the group permissions and gid if the mmap-group option is specified
            mmap_group = (mmap_group or "")
            if POSIX and mmap_group and mmap_group not in FALSE_OPTIONS:
                if mmap_group == "SOCKET":
                    group_id = get_socket_group(socket_filename)
                elif mmap_group.lower() == "auto":
                    group_id = xpra_group()
                    if not group_id and socket_filename:
                        group_id = get_socket_group(socket_filename)
                else:
                    group_id = get_group_id(mmap_group)
                if group_id > 0:
                    log("setting mmap file %s to group id=%i", mmap_filename, group_id)
                    try:
                        os.fchown(fd, -1, group_id)
                    except OSError as e:
                        log("fchown(%i, %i, %i) on %s", fd, -1, group_id, mmap_filename, exc_info=True)
                        log.error("Error: failed to change group ownership of mmap file to '%s':", mmap_group)
                        log.estr(e)
                    from stat import S_IRUSR, S_IWUSR, S_IRGRP, S_IWGRP
                    os.fchmod(fd, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)
            log("using mmap file %s, fd=%s, size=%s", mmap_filename, fd, mmap_size)
            os.lseek(fd, mmap_size - 1, os.SEEK_SET)
            assert os.write(fd, b'\x00')
            os.lseek(fd, 0, os.SEEK_SET)
            mmap_area = mmap.mmap(fd, length=mmap_size)
            if MADVISE:
                madvise(mmap_area)
        return True, delete, mmap_area, mmap_size, mmap_temp_file, mmap_filename
    except Exception as e:
        log("failed to setup mmap: %s", e, exc_info=True)
        log.error("Error: mmap setup failed:")
        log.estr(e)
        if delete:
            if mmap_temp_file:
                try:
                    mmap_temp_file.close()
                except OSError:
                    log("%s()", mmap_temp_file.close, exc_info=True)
                    log.error(" failed to remove the mmap temp file: %s", e)
            else:
                clean_mmap(mmap_filename)
        return rerr()


def clean_mmap(mmap_filename: str) -> None:
    log("clean_mmap(%s)", mmap_filename)
    if mmap_filename and os.path.exists(mmap_filename):
        try:
            os.unlink(mmap_filename)
        except OSError as e:
            log.error("Error: failed to remove the mmap file '%s':", mmap_filename)
            log.estr(e)


def write_mmap_token(mmap_area, token, index: int, count: int = DEFAULT_TOKEN_BYTES) -> None:
    assert count > 0
    # write the token one byte at a time - no endianness
    log("write_mmap_token(%s, %#x, %#x, %#x)", mmap_area, token, index, count)
    v = token
    for i in range(0, count):
        poke = c_ubyte.from_buffer(mmap_area, index + i)
        poke.value = v % 256
        v = v >> 8
    assert v == 0, "token value is too big"


def read_mmap_token(mmap_area, index: int, count: int = DEFAULT_TOKEN_BYTES) -> int:
    assert count > 0
    v = 0
    for i in range(0, count):
        v = v << 8
        peek = c_ubyte.from_buffer(mmap_area, index + count - 1 - i)
        v += peek.value
    log("read_mmap_token(%s, %#x, %#x)=%#x", mmap_area, index, count, v)
    return v


def init_server_mmap(mmap_filename: str, mmap_size: int = 0) -> tuple[Any | None, int]:
    """
        Reads the mmap file provided by the client
        and verifies the token if supplied.
        Returns the mmap object and its size: (mmap, size)
    """
    mmap_area = None
    try:
        import mmap
        if not WIN32:
            try:
                f = open(mmap_filename, "r+b")
            except Exception as e:
                log.error(f"Error: cannot access mmap file {mmap_filename!r}:")
                log.estr(e)
                log.error(" see mmap-group option?")
                return None, 0
            actual_mmap_size = os.path.getsize(mmap_filename)
            if mmap_size and actual_mmap_size != mmap_size:
                log.warn("Warning: expected mmap file '%s' of size %i but got %i",
                         mmap_filename, mmap_size, actual_mmap_size)
            mmap_area = mmap.mmap(f.fileno(), mmap_size)
            f.close()
            return mmap_area, actual_mmap_size
        assert sys.platform == "win32"
        if mmap_size == 0:
            log.error("Error: client did not supply the mmap area size")
            log.error(" try updating your client version?")
            return None, 0
        mmap_area = mmap.mmap(0, mmap_size, mmap_filename)
        return mmap_area, mmap_size
    except Exception:
        log.error("Error: cannot use mmap file '%s'", mmap_filename, exc_info=True)
        if mmap_area:
            mmap_area.close()
        return None, 0


def int_from_buffer(mmap_area, pos: int) -> c_uint32:
    return c_uint32.from_buffer(mmap_area, pos)  # @UndefinedVariable


# descr_data is a list of (offset, length)
# areas from the mmap region
def mmap_read(mmap_area, *descr_data: tuple[int, int]) -> tuple[bytes | memoryview, PaintCallback]:
    """
        Reads data from the mmap_area as written by 'mmap_write'.
        The descr_data is the list of mmap chunks used.
    """
    data_start: c_uint32 = int_from_buffer(mmap_area, 0)
    mv = memoryview(mmap_area)
    if len(descr_data) == 1:
        # construct a zero copy buffer directly from the mmap zone
        # we can only move the `data_start` shared pointer after the buffer has been used
        offset, length = descr_data[0]
        end = offset + length

        def free_mem(*_args) -> None:
            data_start.value = end

        return mv[offset:end], free_mem
    # re-construct the buffer from discontiguous chunks
    # and concatenate them into a byte buffer
    data = []
    end = data_start.value
    for offset, length in descr_data:
        end = offset + length
        data.append(mv[offset:end])
    bdata = b"".join(data)
    # we can not update the shared pointer:
    data_start.value = end
    return bdata, noop


def mmap_write(mmap_area, mmap_size: int, data) -> list[tuple[int, int]]:
    """
        Sends 'data' to the client via the mmap shared memory region,
        returns the chunks of the mmap area used (or None if it failed)
        and the mmap area's free memory.
    """
    size = len(data)
    if size > (mmap_size - 8):
        log.warn("Warning: mmap area is too small!")
        log.warn(" we need to store %s bytes but the mmap area is limited to %i", size, (mmap_size - 8))
        return []
    # This is best explained using diagrams:
    # mmap_area=[&S&E-------------data-------------]
    # The first pair of 4 bytes are occupied by:
    # S=data_start index is only updated by the client and tells us where it has read up to
    # E=data_end index is only updated here and marks where we have written up to (matches current seek)
    # '-' denotes unused/available space
    # '+' is for data we have written
    # '*' is for data we have just written in this call
    # E and S show the location pointed to by data_start/data_end
    mmap_data_start: c_uint32 = int_from_buffer(mmap_area, 0)
    mmap_data_end: c_uint32 = int_from_buffer(mmap_area, 4)
    start = max(8, mmap_data_start.value)
    end = max(8, mmap_data_end.value)
    log("mmap: start=%i, end=%i, size of data to write=%i", start, end, size)
    if end < start:
        # we have wrapped around but the client hasn't yet:
        # [++++++++E--------------------S+++++]
        # so there is one chunk available (from E to S) which we will use:
        # [++++++++************E--------S+++++]
        available = start - end
        chunk = available
    else:
        # we have not wrapped around yet, or the client has wrapped around too:
        # [------------S++++++++++++E---------]
        # so there are two chunks available (from E to the end, from the start to S):
        # [****--------S++++++++++++E*********]
        chunk = mmap_size - end
        available = chunk + (start - 8)
    if size > available:
        log.warn("Warning: mmap area is full!")
        log.warn(" we need to store %s bytes but only have %s free space left", size, available)
        return []
    if size < chunk:
        # data fits in the first chunk:
        # ie: initially:
        # [----------------------------------]
        # [*********E------------------------]
        # or if data already existed:
        # [+++++++++E------------------------]
        # [+++++++++**********E--------------]
        mmap_area.seek(end)
        mmap_area.write(data)
        chunks = [(end, size)]
        mmap_data_end.value = end + size
    else:
        # data does not fit in first chunk alone:
        if not ALWAYS_WRAP and available >= (mmap_size / 2) and available >= (size * 3) and size < (start - 8):
            # still plenty of free space, don't wrap around: just start again:
            # [------------------S+++++++++E------]
            # [*******E----------S+++++++++-------]
            mmap_area.seek(8)
            mmap_area.write(data)
            chunks = [(8, size)]
            mmap_data_end.value = 8 + size
        else:
            # split in 2 chunks: wrap around the end of the mmap buffer:
            # [------------------S+++++++++E------]
            # [******E-----------S+++++++++*******]
            mmap_area.seek(end)
            mmap_area.write(data[:chunk])
            mmap_area.seek(8)
            mmap_area.write(data[chunk:])
            l2 = size - chunk
            chunks = [(end, chunk), (8, l2)]
            mmap_data_end.value = 8 + l2
    log("mmap_write: %s bytes", len(data))
    return chunks


def mmap_free_size(mmap_area, mmap_size: int) -> int:
    mmap_data_start: c_uint32 = int_from_buffer(mmap_area, 0)
    mmap_data_end: c_uint32 = int_from_buffer(mmap_area, 4)
    start = max(8, mmap_data_start.value)
    end = min(mmap_size, max(8, mmap_data_end.value))
    if end < start:
        # we have wrapped around but the client hasn't yet:
        # [++++++++E--------------------S+++++]
        # so there is one chunk available (from E to S)
        return start - end
    # we have not wrapped around yet, or the client has wrapped around too:
    # [------------S++++++++++++E---------]
    # so there are two chunks available (from E to the end, from the start to S)
    return (start - 8) + (mmap_size - end)


class BaseMmapArea:
    """
    Represents an mmap area we can read from or write to
    """

    def __init__(self, name: str, filename="", size=0):
        self.name = name
        self.mmap = None
        self.enabled = False
        self.token: int = 0
        self.token_index: int = 0
        self.token_bytes: int = 0
        self.filename = filename
        self.size = size

    def __repr__(self):
        return "MmapArea(%s:%s:%i)" % (self.name, self.filename, self.size)

    def __bool__(self):
        return bool(self.mmap) and self.enabled and self.size > 0

    def close(self):
        mmap = self.mmap
        if mmap:
            try:
                mmap.close()
            except BufferError:
                log("%s.close()", mmap, exc_info=True)
            except OSError:
                log("%s.close()", mmap, exc_info=True)
                log.warn("Warning: failed to close %s mmap area", self.name)
            self.mmap = None
        self.enabled = False

    def get_caps(self) -> dict[str, Any]:
        return {
            "file": self.filename,
            "size": self.size,
            "token": self.token,
            "token_index": self.token_index,
            "token_bytes": self.token_bytes,
        }

    def get_info(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "file": self.filename,
            "size": self.size,
        }

    def parse_caps(self, mmap_caps: typedict) -> None:
        self.enabled = mmap_caps.boolget("enabled", True)
        self.token = mmap_caps.intget("token")
        self.token_index = mmap_caps.intget("token_index", 0)
        self.token_bytes = mmap_caps.intget("token_bytes", DEFAULT_TOKEN_BYTES)
        self.size = self.size or mmap_caps.intget("size")

    def verify_token(self) -> bool:
        if not self.mmap:
            raise RuntimeError("mmap object is not defined")
        token = read_mmap_token(self.mmap, self.token_index, self.token_bytes)
        if token == 0:
            log.info(f"the server is not using the {self.name!r} mmap area")
            return False
        if token != self.token:
            self.enabled = False
            log.error(f"Error: {self.name!r} mmap token verification failed!")
            log.error(f" expected {self.token:x}")
            log.error(f" found {token:x}")
            log.error(" mmap is disabled")
            return False
        log.info("enabled fast %s mmap transfers using %sB shared memory area",
                 self.name, std_unit(self.size, unit=1024))
        return True

    def gen_token(self) -> None:
        self.token = get_int_uuid()
        self.token_bytes = DEFAULT_TOKEN_BYTES
        self.token_index = randint(0, self.size - DEFAULT_TOKEN_BYTES)

    def write_token(self) -> None:
        write_mmap_token(self.mmap, self.token, self.token_index, self.token_bytes)

    def write_data(self, data) -> list[tuple[int, int]]:
        return mmap_write(self.mmap, self.size, data)

    def mmap_read(self, *descr_data: tuple[int, int]) -> tuple[bytes | memoryview, PaintCallback]:
        return mmap_read(self.mmap, *descr_data)

    def get_free_size(self) -> int:
        return mmap_free_size(self.mmap, self.size)



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from io import StringIO
from contextlib import redirect_stderr
from xpra.x11.bindings.xlib cimport (
    Success,
    Display, Window, Atom, Status,
    XInternAtom,
    XOpenDisplay, XCloseDisplay,
    XQueryExtension, XFree,
    XGetWindowProperty, XDefaultRootWindow,
)
from xpra.x11.bindings.randr cimport get_monitor_properties
from xpra.log import Logger

log = Logger("x11")

DEF XNone = 0


def isX11(display_name: str=os.environ.get("DISPLAY", "")) -> bool:
    b = display_name.encode()
    cdef char* display = b
    cdef Display *d = NULL
    with redirect_stderr(StringIO()) as f:
        d = XOpenDisplay(display)
    if not d:
        log(f"isX11({display_name}) cannot open display: %s", f.getvalue())
        return False
    return True


def isxwayland(display_name: str=os.environ.get("DISPLAY", "")) -> bool:
    b = display_name.encode()
    cdef char* display = b
    cdef Display *d = NULL
    with redirect_stderr(StringIO()) as f:
        d = XOpenDisplay(display)
    if not d:
        log(f"isxwayland({display_name}) cannot open display: %s", f.getvalue())
        return False
    cdef int opcode, event, error
    try:
        #the easy way:
        if XQueryExtension(d, "XWAYLAND", &opcode, &event, &error):
            log(f"isxwayland({display_name}) XWAYLAND extension found")
            return True
        #surely a vfb is not wayland?
        if get_xstring(d, "VFB_IDENT"):
            log(f"isxwayland({display_name}) VFB_IDENT found")
            return False
        #this can go wrong...
        props = get_monitor_properties(d)
        log(f"isxwayland({display_name}) monitor properties={props}")
        for mprops in props.values():
            if mprops.get("name", "").startswith("XWAYLAND"):
                log(f"isxwayland({display_name}) found XWAYLAND monitor: {mprops}")
                return True
        return False
    finally:
        XCloseDisplay(d)


cdef Atom intern_atom(Display *display, name: str):
    b = name.encode()
    return XInternAtom(display, b, True)


cdef str get_xstring(Display *display, name="XPRA_SERVER_UUID", prop_type="STRING"):
    cdef Atom xactual_type = <Atom> 0
    cdef int actual_format = 0
    cdef unsigned long nitems = 0, bytes_after = 0
    cdef unsigned char * prop_data = <unsigned char*> 0
    cdef Atom xtype = 0, prop = 0

    def fail(message):
        #print(message)
        return ""

    xtype = intern_atom(display, prop_type)
    if xtype==0:
        #the property cannot exist if the type is not defined
        return fail(f"Atom {prop_type} does not exist")
    prop = intern_atom(display, name)
    if prop==0:
        #the atom for the name does not exist so the property cannot exist:
        return fail(f"Atom {name} does not exist")
    cdef Window root = XDefaultRootWindow(display)
    cdef int buffer_size = 128
    cdef Status status = XGetWindowProperty(display, root, prop,
                                            0,
                                            buffer_size // 4,
                                            False,
                                            xtype, &xactual_type,
                                            &actual_format, &nitems, &bytes_after, &prop_data)
    if status != Success:
        return fail(f"XGetWindowProperty failed")
    if xactual_type == XNone:
        return fail(f"no actual type")
    if xtype != xactual_type:
        return fail(f"actual type differs")
    if nitems==0:
        return fail(f"no data")
    if bytes_after:
        return fail(f"more data than expected")
    if actual_format == 8:
        bytes_per_item = 1
    elif actual_format == 16:
        bytes_per_item = sizeof(short)
    elif actual_format == 32:
        bytes_per_item = sizeof(long)
    else:
        raise RuntimeError(f"unexpected format value: {actual_format}")
    cdef int nbytes = bytes_per_item * nitems
    data = (<char *> prop_data)[:nbytes]
    return data.decode("latin1")



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable, Sequence

from xpra.x11.bindings.xlib cimport (
    Display, Time, Bool, Status, Atom,
    XInternAtom, XInternAtoms,
    XGetAtomName,
    XFree,
    XGetErrorText,
    XUngrabKeyboard, XUngrabPointer,
    XSynchronize, XSync, XFlush,
    CurrentTime,
    XDefaultRootWindow,
)
from libc.stdlib cimport malloc, free        # pylint: disable=syntax-error
from libc.stdint cimport uintptr_t

from xpra.util.env import envbool
from xpra.util.str_fn import strtobytes
from xpra.util.system import is_X11
from xpra.log import Logger
log = Logger("x11", "bindings", "core")

from xpra.x11.bindings.display_source cimport get_display
from xpra.x11.bindings.display_source import get_display_name


def noop(*args) -> None:
    pass


cdef object context_check = noop


def set_context_check(fn: Callable) -> None:
    global context_check
    context_check = fn


def call_context_check(*args) -> None:
    context_check(*args)


cdef class X11CoreBindingsInstance:

    def __cinit__(self):
        if not is_X11():
            raise RuntimeError("cannot load X11 bindings with wayland")
        self.display = get_display()
        if self.display == NULL:
            raise RuntimeError("X11 display is not set")
        bstr = get_display_name().encode("latin1")
        self.display_name = bstr
        self.XSynchronize(envbool("XPRA_X_SYNC", False))

    def XSynchronize(self, enable: bool) -> None:
        XSynchronize(self.display, enable)

    def XSync(self, discard=False) -> None:
        XSync(self.display, discard)

    def XFlush(self) -> None:
        XFlush(self.display)

    def context_check(self, *args) -> None:
        global context_check
        context_check(*args)

    def __repr__(self):
        return "X11CoreBindings(%s)" % self.display_name

    def get_root_xid(self) -> long:
        assert self.display
        return XDefaultRootWindow(self.display)

    cdef Atom str_to_atom(self, atomstr):
        """Returns the X atom corresponding to the given Python string or Python
        integer (assumed to already be an X atom)."""
        self.context_check("str_to_atom")
        bstr = strtobytes(atomstr)
        cdef char* string = bstr
        assert self.display!=NULL, "display is closed"
        return XInternAtom(self.display, string, False)

    cdef Atom xatom(self, str_or_int):
        """Returns the X atom corresponding to the given Python string or Python
        integer (assumed to already be an X atom)."""
        if isinstance(str_or_int, int):
            return <Atom> str_or_int
        return self.str_to_atom(str_or_int)

    def intern_atoms(self, atom_names: Sequence[str]) -> None:
        cdef int count = len(atom_names)
        cdef char** names = <char **> malloc(sizeof(uintptr_t)*(count+1))
        assert names!=NULL
        cdef Atom* atoms_return = <Atom*> malloc(sizeof(Atom)*(count+1))
        assert atoms_return!=NULL
        from ctypes import create_string_buffer, addressof
        str_names = [create_string_buffer(x.encode("latin1")) for x in atom_names]
        cdef uintptr_t ptr
        for i, x in enumerate(str_names):
            ptr = addressof(x)
            names[i] = <char*> ptr
        cdef Status s = XInternAtoms(self.display, names, count, 0, atoms_return)
        free(names)
        free(atoms_return)
        assert s!=0, "failed to intern some atoms"

    def get_xatom(self, str_or_int) -> Atom:
        return self.xatom(str_or_int)

    def XGetAtomName(self, Atom atom) -> bytes:
        self.context_check("XGetAtomName")
        cdef char *v = XGetAtomName(self.display, atom)
        if v == NULL:
            return b""
        r = v[:]
        XFree(v)
        return r

    def get_atom_name(self, Atom atom) -> str:
        bin_name = self.XGetAtomName(atom)
        return bin_name.decode("latin1")

    def get_error_text(self, code) -> str:
        if self.display == NULL:
            raise RuntimeError("display is closed")
        if not isinstance(code, int):
            return str(code)
        cdef char[128] buffer
        XGetErrorText(self.display, code, buffer, 128)
        return (bytes(buffer[:128]).split(b"\0", 1)[0]).decode("latin1")

    def UngrabKeyboard(self, Time time=CurrentTime) -> long:
        self.context_check("UngrabKeyboard")
        if self.display == NULL:
            raise RuntimeError("display is closed")
        return XUngrabKeyboard(self.display, time)

    def UngrabPointer(self, Time time=CurrentTime) -> long:
        self.context_check("UngrabPointer")
        if self.display == NULL:
            raise RuntimeError("display is closed")
        return XUngrabPointer(self.display, time)


cdef X11CoreBindingsInstance singleton = None


def X11CoreBindings() -> X11CoreBindingsInstance:
    global singleton
    if singleton is None:
        singleton = X11CoreBindingsInstance()
    return singleton



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import uuid
from queue import Queue
from typing import Any
from time import monotonic
from collections import deque
from collections.abc import Sequence
from threading import Event

from xpra import __version__
from xpra.common import FULL_INFO
from xpra.os_util import gi_import
from xpra.util.str_fn import Ellipsizer
from xpra.util.io import is_socket, load_binary_file
from xpra.util.env import envint, osexpand
from xpra.util.objects import typedict, AtomicInteger
from xpra.util.thread import start_thread
from xpra.scripts.main import connect_to
from xpra.scripts.parsing import parse_display_name
from xpra.scripts.config import InitExit, make_defaults_struct
from xpra.net.common import PacketElement, PacketType
from xpra.codecs.constants import VideoSpec, TransientCodecException
from xpra.log import Logger

log = Logger("remote")

GLib = gi_import("GLib")

ENCODER_SERVER_TIMEOUT = envint("XPRA_ENCODER_SERVER_TIMEOUT", 5)
ENCODER_SERVER_SOCKET_TIMEOUT = envint("XPRA_ENCODER_SERVER_SOCKET_TIMEOUT", 1)
CONNECT_POLL_DELAY = envint("XPRA_CONNECT_POLL_DELAY", 10000)
RECONNECT_DELAY = envint("XPRA_RECONNECT_DELAY", 2000)


try:
    from xpra.client.mixins.mmap import MmapClient
    baseclass = MmapClient
except ImportError:
    from xpra.client.base.stub_client_mixin import StubClientMixin
    baseclass = StubClientMixin


def get_version() -> Sequence[int]:
    return 0, 1


def get_type() -> str:
    return "remote"


def get_info() -> dict[str, Any]:
    return {"version": get_version()}


def find_server_uri(sessions_dir: str) -> str:
    session_dir = os.environ.get("XPRA_SESSION_DIR")
    log(f"find_server_uri({sessions_dir!r}) {session_dir=!r}")
    encoder_sockets = []
    for sdir in os.listdir(sessions_dir):
        if not sdir.isnumeric():
            continue
        path = os.path.join(sessions_dir, sdir)
        if path == session_dir:
            continue
        encoder_socket = get_server_socket_path(path)
        if encoder_socket:
            encoder_sockets.append(encoder_socket)
    if not encoder_sockets:
        return ""
    if len(encoder_sockets) > 1:
        # sort them by last modified time:
        times = {}
        for spath in encoder_sockets:
            times[spath] = os.path.getmtime(spath)
        # sort by value:
        times = dict(sorted(times.items(), key=lambda item: - item[1]))
        encoder_sockets = tuple(times.keys())
    encoder = encoder_sockets[0]
    log(f"find_server_uri({sessions_dir!r})=%s, from %i encoders found: %s",
        encoder,len(encoder_sockets), encoder_sockets)
    return encoder


SYSTEM_ENCODER_PATH = "/run/xpra/encoder"


def get_server_socket_path(session_dir: str) -> str:
    if os.name == "posix" and os.path.exists(SYSTEM_ENCODER_PATH) and is_socket(SYSTEM_ENCODER_PATH):
        return SYSTEM_ENCODER_PATH
    if not os.path.exists(session_dir) or not os.path.isdir(session_dir):
        return ""
    # there must be a server config:
    config = os.path.join(session_dir, "config")
    if not os.path.exists(config) or not os.path.isfile(config):
        return ""
    # and a socket we can connect to:
    socket = os.path.join(session_dir, "socket")
    if not os.path.exists(socket) or not is_socket(socket):
        return ""
    # verify that the server is an 'encoding' server:
    cdata = load_binary_file(config)
    if not cdata:
        return ""
    for line in cdata.decode("utf8").splitlines():
        if line.startswith("mode="):
            mode = line[len("mode="):]
            if mode == "encoder":
                return socket
    return ""


class RemoteConnectionClient(baseclass):

    def __init__(self, options: dict):
        super().__init__()
        log(f"RemoteConnectionClient({options})")
        to = typedict(options)
        self.uri = to.strget("uri", "")
        self.server_timeout = to.intget("timeout", ENCODER_SERVER_TIMEOUT)
        self.server_socket_timeout = to.intget("socket-timeout", ENCODER_SERVER_SOCKET_TIMEOUT)
        self.reconnect_delay = to.intget("reconnect-delay", RECONNECT_DELAY)
        self.connect_poll_delay = to.intget("connect-poll-delay", CONNECT_POLL_DELAY)
        self.lz4 = to.boolget("lz4", False)
        self.connect_timer = 0
        self.connecting = False
        self.specs: dict[str, dict[str, Sequence[VideoSpec]]] = {}
        self.protocol = None
        self._ordinary_packets = []
        self.connect_event = Event()
        opts = make_defaults_struct()
        opts.mmap = "both"
        opts.mmap_group = ""
        super().init(opts)
        self.sessions_dir = osexpand(to.strget("sessions-dir", opts.sessions_dir))

    def cleanup(self) -> None:
        self.cancel_schedule_connect()
        super().cleanup()

    def cancel_schedule_connect(self) -> None:
        ct = self.connect_timer
        if ct:
            self.connect_timer = 0
            GLib.source_remove(ct)

    def schedule_connect(self, delay: int) -> None:
        connected = self.is_connected()
        log(f"schedule_connect({delay}) {connected=}, timer={self.connect_timer}, connecting={self.connecting}")
        if connected or self.connect_timer or self.connecting:
            return
        self.connect_timer = GLib.timeout_add(delay, self.scheduled_connect, delay)

    def scheduled_connect(self, delay: int) -> bool:
        self.connect_timer = 0
        # connect() does I/O, so we have to use a separate thread to call it:
        start_thread(self.threaded_scheduled_connect, "encoder-server-connect", True, (delay, ))
        return False

    def threaded_scheduled_connect(self, delay: int) -> None:
        if self.is_connected() or self.connecting:
            return
        if not self.do_connect():
            # try again:
            self.schedule_connect(delay)

    def connect(self) -> None:
        if self.is_connected() or self.connecting:
            log("already connected / connecting")
            return
        if self.connect_timer:
            log("connect timer is already due")
            return
        if not self.do_connect():
            self.schedule_connect(self.connect_poll_delay)

    def do_connect(self) -> bool:
        uri = self.uri or find_server_uri(self.sessions_dir)
        if not uri:
            log("no encoder server found")
            return False
        try:
            self.connecting = True
            opts = make_defaults_struct()
            desc = parse_display_name(ValueError, opts, uri)
            if "timeout" not in desc:
                desc["timeout"] = self.server_socket_timeout
            desc["quiet"] = True
            log(f"EncoderClient.do_connect() server desc={desc!r}")
            conn = connect_to(desc, opts)
            super().setup_connection(conn)
            self.protocol = self.make_protocol(conn)
            self.send_hello()
        except InitExit as e:
            log("failed to connect: %s", e)
            return False
        except OSError:
            log("failed to connect", exc_info=True)
            return False
        finally:
            self.connecting = False
        return True

    def make_protocol(self, conn):
        from xpra.net.packet_encoding import init_all
        init_all()
        from xpra.net.compression import init_all
        init_all()
        from xpra.net.protocol.factory import get_client_protocol_class
        protocol_class = get_client_protocol_class(conn.socktype)
        protocol = protocol_class(conn, self._process_packet, self._next_packet)
        protocol.enable_default_encoder()
        protocol.enable_default_compressor()
        protocol._log_stats = False
        protocol.large_packets += ["encodings", "context-compress", "context-data"]
        # self.add_packet_handler("setting-change", noop)
        # if conn.timeout > 0:
        #    GLib.timeout_add((conn.timeout + EXTRA_TIMEOUT) * 1000, self.verify_connected)
        return protocol

    def is_connected(self) -> bool:
        p = self.protocol
        return bool(p) and not p.is_closed()

    def send(self, packet_type: str, *parts: PacketElement) -> None:
        packet = (packet_type, *parts)
        self._ordinary_packets.append(packet)
        self.protocol.source_has_more()

    def _process_packet(self, proto, packet: PacketType) -> None:
        packet_type = packet[0]
        if packet_type in (
                "hello", "encodings", "startup-complete",
                "setting-change",
                "connection-lost", "disconnect",
                "context-response", "context-data",
        ):
            fn = getattr(self, "_process_%s" % packet_type.replace("-", "_"))
            fn(packet)
        else:
            log.warn(f"Warning: received unexpected {packet_type!r} from encoder server connection {proto}")

    def send_hello(self) -> None:
        caps = {
            "version": __version__[:FULL_INFO+1],
            "client_type": "encode",
            "uuid": uuid.uuid4().hex,
            "windows": False,
            "keyboard": False,
            "wants": ("encodings", "video", ),
            "encoding": {"core": ("rgb32", "rgb24", )},
            "mouse": False,
            "network-state": False,  # tell older server that we don't have "ping"
        }
        caps.update(super().get_caps())
        from xpra.net.packet_encoding import get_packet_encoding_caps
        caps.update(get_packet_encoding_caps(0))
        from xpra.net.compression import get_compression_caps
        caps.update(get_compression_caps(0))
        log(f"sending hello={caps!r}")
        self.send("hello", caps)
        self.connect_event.clear()
        self.protocol.start()
        self.connect_event.wait(self.server_timeout)

    def _process_hello(self, packet: PacketType) -> None:
        caps = packet[1]
        log("got hello: %s", Ellipsizer(caps))
        if not super().parse_server_capabilities(typedict(caps)):
            raise RuntimeError("failed to parse capabilities")
        version = caps.get("version")
        log.info("connected to encoder server version %s", version)

    def _process_disconnect(self, packet: PacketType) -> None:
        log("disconnected from server %s", self.protocol)
        self.server_connection_cleanup()

    def _process_connection_lost(self, packet: PacketType) -> None:
        log("connection-lost for server %s", self.protocol)
        self.server_connection_cleanup()

    def server_connection_cleanup(self):
        self.protocol = None
        super().cleanup()
        self.schedule_connect(self.reconnect_delay)

    def _process_startup_complete(self, packet: PacketType) -> None:
        log(f"{packet!r}")
        self.connect_event.set()

    def _next_packet(self) -> tuple[Any, bool, bool]:
        return self._ordinary_packets.pop(0), True, bool(self._ordinary_packets)

    def disconnect(self) -> None:
        p = self.protocol
        if p:
            self.protocol = None
            p.close()


class RemoteCodec:
    __slots__ = (
        "server", "codec_type",
        "generation", "encoding",
        "width", "height", "pixel_format", "last_frame_times",
        "ready", "closed", "responses",
        "__weakref__",
    )
    """
    This encoder connects to an encoder server and delegates to it
    """
    counter = AtomicInteger()

    def __init__(self, server: RemoteConnectionClient, codec_type: str):
        self.server = server
        self.codec_type = codec_type
        self.generation = self.counter.increase()
        self.encoding = ""
        self.closed = False
        self.width = self.height = 0
        self.pixel_format = ""
        self.last_frame_times: deque[float] = deque(maxlen=200)
        self.ready = False
        self.responses = Queue(maxsize=1)

    def init_context(self, encoding: str, width: int, height: int, pixel_format: str, options: typedict) -> None:
        self.encoding = encoding
        self.width = width
        self.height = height
        self.pixel_format = pixel_format
        self.closed = False
        try:
            self.server.connect()
        except (OSError, InitExit) as e:
            log("failed to connect to remote encoder server %s", self.server)
            log(" %s", e)
            raise TransientCodecException(f"failed to connect: {e}") from None
        if not self.server.is_connected():
            raise TransientCodecException("not connected")

    def is_ready(self) -> bool:
        return self.ready

    def get_info(self) -> dict[str, Any]:
        info = get_info()
        if not self.pixel_format:
            return info
        info.update({
            "width": self.width,
            "height": self.height,
            "encoding": self.encoding,
            "pixel-format": self.pixel_format,
        })
        # calculate fps:
        now = monotonic()
        last_time = now
        cut_off = now - 10.0
        f = 0
        for v in tuple(self.last_frame_times):
            if v > cut_off:
                f += 1
                last_time = min(last_time, v)
        if f > 0 and last_time < now:
            info["fps"] = int(0.5 + f / (now - last_time))
        return info

    def __repr__(self):
        if not self.pixel_format:
            return "remote-codec(uninitialized)"
        return f"remote-codec({self.pixel_format} - {self.width}x{self.height})"

    def is_closed(self) -> bool:
        return self.closed

    def get_encoding(self) -> str:
        return self.encoding

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_type(self) -> str:
        return f"remote-{self.codec_type}"

    def clean(self) -> None:
        self.closed = True
        self.width = 0
        self.height = 0
        self.encoding = ""
        self.pixel_format = ""
        self.last_frame_times = deque()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.util.str_fn import csv
from xpra.util.env import envbool
from xpra.common import noop
from xpra.gtk.widget import scaled_image
from xpra.gtk.dialogs.about import about
from xpra.client.gtk3.tray_menu import (
    GTKTrayMenu,
    CLIPBOARD_LABEL_TO_NAME, CLIPBOARD_NAME_TO_LABEL, CLIPBOARD_LABELS,
    CLIPBOARD_DIRECTION_LABELS, CLIPBOARD_DIRECTION_NAME_TO_LABEL,
    SHOW_UPLOAD, SHOW_VERSION_CHECK, RUNCOMMAND_MENU, SHOW_SERVER_COMMANDS, SHOW_SHUTDOWN,
    SHOW_QR,
)
from xpra.platform.paths import get_icon
from xpra.platform.darwin import get_OSXApplication
from xpra.client.base import features
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("osx", "tray", "menu")
clipboardlog = Logger("osx", "menu", "clipboard")

# control which menus are shown in the OSX global menu:
DEFAULT = envbool("XPRA_OSX_SHOW_MENU_DEFAULT", True)
SHOW_FEATURES_MENU = DEFAULT
SHOW_SOUND_MENU = DEFAULT
SHOW_ENCODINGS_MENU = DEFAULT
SHOW_ACTIONS_MENU = DEFAULT
SHOW_INFO_MENU = DEFAULT
SHOW_CLIPBOARD_MENU = DEFAULT
SHOW_SERVER_MENU = DEFAULT

SHOW_ABOUT_XPRA = DEFAULT

SINGLE_MENU = envbool("XPRA_OSX_SINGLE_MENU", False)
USE_WINDOW_MENU = envbool("XPRA_OSX_USE_WINDOW_MENU", True)

SEPARATOR = "SEPARATOR"

_OSXMenuHelper = None


def getOSXMenuHelper(client=None):
    global _OSXMenuHelper
    if _OSXMenuHelper is None:
        _OSXMenuHelper = OSXMenuHelper(client)
    elif client is not None:
        _OSXMenuHelper.set_client(client)
    return _OSXMenuHelper


class OSXMenuHelper(GTKTrayMenu):
    """
    we have to do this stuff here,
    so we can re-use the same instance,
    and change the callbacks if needed.
    (that way, the launcher and the client can both change the menus)
    """

    def __init__(self, client=None):
        super().__init__(client)
        log("OSXMenuHelper(%s)", client)
        self.menu_bar = None
        self.hidden_window = None
        self.menus = {}  # all the "top-level" menu items we manage
        self.app_menus = {}  # the ones added to the app_menu via insert_app_menu_item (which cannot be removed!)
        self.dock_menu = None
        self.window_menu = None
        self.window_menu_item = None
        self.numlock_menuitem = None
        self.full = False
        self.set_client(client)
        self._clipboard_change_pending = False

    def set_client(self, client) -> None:
        super().set_client(client)
        # if we call add_about before the main loop is ready,
        # things don't work...
        if client and SHOW_ABOUT_XPRA:
            self.after_handshake(self.add_about)

    def show_menu(self, button: int, time) -> None:
        # does not mean anything on OSX since the menu is controlled by the OS
        pass

    def build(self):
        log("OSXMenuHelper.build()")
        if self.menu_bar is None:
            self.menu_bar = Gtk.MenuBar()
            self.menu_bar.show_all()
        return self.menu_bar

    def rebuild(self):
        log("OSXMenuHelper.rebuild()")
        if not self.menu_bar:
            return self.build()
        self.remove_all_menus()
        return self.menu_bar

    def remove_all_menus(self) -> None:
        log("OSXMenuHelper.remove_all_menus()")
        if self.menu_bar:
            for x in self.menus.values():
                if x in self.menu_bar.get_children():
                    self.menu_bar.remove(x)
                    x.hide()
        self.menus = {}
        self.full = False

    def add_top_level_menu(self, label: str, submenu) -> None:
        """ Adds the item to the app-menu or to the top bar,
            but only if it has not been added yet.
            (depending on the SINGLE_MENU flag)
        """
        if SINGLE_MENU:
            # add or re-use menu item in the app-menu:
            self.add_to_app_menu(label, submenu)
        else:
            self.add_to_menu_bar(label, submenu)

    def add_to_app_menu(self, label: str, submenu) -> None:
        item = self.app_menus.get(label)
        if item:
            log("application menu already has a '%s' entry", label)
            if submenu is not None:
                item.set_submenu(submenu)
            item.show_all()
        else:
            if label.startswith(SEPARATOR):
                item = Gtk.SeparatorMenuItem()
            else:
                item = self.menuitem(label)
                item.set_submenu(submenu)
            item.show_all()
            macapp = get_OSXApplication()
            macapp.insert_app_menu_item(item, 1)
            self.app_menus[label] = item

    def add_to_menu_bar(self, label, submenu):
        if label.startswith(SEPARATOR):
            return  # not relevant
        item = self.menus.get(label)
        if item is None:
            item = self.menuitem(label)
            item.set_submenu(submenu)
            item.show_all()
            self.menu_bar.add(item)
            self.menus[label] = item
        else:
            item.set_submenu(submenu)
            item.show_all()

    def get_menu(self, label: str):
        if SINGLE_MENU:
            return self.app_menus.get(label)
        return self.menus.get(label)

    def add_about(self) -> None:
        if "About" in self.app_menus:
            return
        item = self.menuitem("About", cb=about)
        item.show_all()
        macapp = get_OSXApplication()
        macapp.insert_app_menu_item(item, 0)
        self.app_menus["About"] = item

    def add_full_menu(self) -> None:
        log("OSXMenuHelper.add_full_menu()")
        if self.full or not self.client:
            return
        self.full = True
        menus = self.get_extra_menus()
        for label, submenu in reversed(menus):
            self.add_top_level_menu(label, submenu)
        self.menu_bar.show_all()
        if USE_WINDOW_MENU:
            self.window_menu_item = self.get_menu("Windows")
            if self.window_menu_item:
                macapp = get_OSXApplication()
                macapp.set_window_menu(self.window_menu_item)

    def get_extra_menus(self) -> list:
        menus = []

        def add(menu, item):
            if item:
                menu.add(item)

        if SHOW_INFO_MENU:
            info_menu = self.make_menu()
            menus.append(("Info", info_menu))
            add(info_menu, self.make_sessioninfomenuitem())
            if SHOW_QR:
                add(info_menu, self.make_qrmenuitem())
            if SHOW_VERSION_CHECK:
                add(info_menu, self.make_updatecheckmenuitem())
            add(info_menu, self.make_bugreportmenuitem())
        if SHOW_FEATURES_MENU:
            features_menu = self.make_menu()
            menus.append(("Features", features_menu))
            self.append_featuresmenuitems(features_menu)
            if features.windows:
                add(features_menu, self.make_swapkeysmenuitem())
                add(features_menu, self.make_invertmousewheelmenuitem())
                add(features_menu, self.make_numlockmenuitem())
                add(features_menu, self.make_scalingmenuitem())
        if features.clipboard and SHOW_CLIPBOARD_MENU:
            clipboard_menu = self.make_menu()
            menus.append(("Clipboard", clipboard_menu))
            for label in CLIPBOARD_LABELS:
                add(clipboard_menu, self.make_clipboard_submenuitem(label, self._remote_clipboard_changed))
            add(clipboard_menu, Gtk.SeparatorMenuItem())
            for label in CLIPBOARD_DIRECTION_LABELS:
                add(clipboard_menu, self.make_clipboard_submenuitem(label, self._clipboard_direction_changed))
            clipboard_menu.show_all()
            self.after_handshake(self.set_clipboard_menu, clipboard_menu)
        if features.audio and SHOW_SOUND_MENU:
            audio_menu = self.make_menu()
            if self.client.speaker_allowed and self.client.speaker_codecs:
                add(audio_menu, self.make_speakermenuitem())
            if self.client.microphone_allowed and self.client.microphone_codecs:
                add(audio_menu, self.make_microphonemenuitem())
            menus.append(("Audio", audio_menu))
        if features.windows and SHOW_ENCODINGS_MENU:
            encodings_menu = self.make_menu()

            def set_encodings_menu(*_args):
                client_encodings, server_encodings = self.get_encoding_options()
                from xpra.client.gtk3.menu_helper import populate_encodingsmenu
                populate_encodingsmenu(encodings_menu, self.get_current_encoding, self.set_current_encoding,
                                       client_encodings, server_encodings)

            self.after_handshake(set_encodings_menu)
            menus.append(("Encoding", encodings_menu))
        if features.windows and SHOW_ACTIONS_MENU:
            actions_menu = self.make_menu()
            add(actions_menu, self.make_raisewindowsmenuitem())
            add(actions_menu, self.make_minimizewindowsmenuitem())
            add(actions_menu, self.make_refreshmenuitem())
            add(actions_menu, self.make_reinitmenuitem())
            self.window_menu = actions_menu
            menus.append(("Windows", actions_menu))
        if RUNCOMMAND_MENU or SHOW_SERVER_COMMANDS or SHOW_UPLOAD or SHOW_SHUTDOWN:
            server_menu = self.make_menu()
            if SHOW_SHUTDOWN:
                add(server_menu, self.make_shutdownmenuitem())

            # set_sensitive(bool) does not work on OSX,
            # so we only add the menu item if it does something
            def add_ah(*_args):
                if self.client.server_start_new_commands:
                    add(server_menu, self.make_runcommandmenuitem())
                if SHOW_SERVER_COMMANDS and self.client.server_commands_info:
                    add(server_menu, self.make_servercommandsmenuitem())
                if SHOW_UPLOAD and self.client.remote_file_transfer:
                    add(server_menu, self.make_uploadmenuitem())

            self.after_handshake(add_ah)
            menus.append(("Server", server_menu))
        menus.append((SEPARATOR + "-EXTRAS", None))
        return menus

    def _clipboard_direction_changed(self, item, label: str):
        clipboardlog("_clipboard_direction_changed(%s, %s) clipboard_change_pending=%s",
                     item, label, self._clipboard_change_pending)
        label = self.select_clipboard_menu_option(item, label, CLIPBOARD_DIRECTION_LABELS)
        self.do_clipboard_direction_changed(label or "")

    def _remote_clipboard_changed(self, item, label: str):
        clipboardlog("_remote_clipboard_changed(%s, %s) clipboard_change_pending=%s",
                     item, label, self._clipboard_change_pending)
        # ensure this is the only clipboard label selected:
        label = self.select_clipboard_menu_option(item, label, CLIPBOARD_LABELS)
        if not label:
            return
        remote_clipboard = CLIPBOARD_LABEL_TO_NAME[label]
        clipboardlog("will select clipboard menu item with label=%s, for remote_clipboard=%s", label, remote_clipboard)
        GLib.timeout_add(0, self._do_clipboard_change, remote_clipboard)

    def _do_clipboard_change(self, remote_clipboard: str):
        # why do we look it up again when we could just pass it in
        # to make_clipboard_submenuitem as an extra argument?
        # because gtk-osx would fall over itself, making a complete mess of the menus in the process
        # and why do we use a timer here? again, more trouble with gtk-osx..
        self._clipboard_change_pending = False
        self.set_new_remote_clipboard(remote_clipboard)

    def make_clipboard_submenuitem(self, label: str, cb: Callable = noop):
        clipboard_item = self.checkitem(label)
        clipboard_item.set_draw_as_radio(True)

        def clipboard_option_changed(item):
            clipboardlog("clipboard_option_changed(%s) label=%s, callback=%s clipboard_change_pending=%s",
                         item, label, cb, self._clipboard_change_pending)
            cb(item, label)

        clipboard_item.connect("toggled", clipboard_option_changed)
        return clipboard_item

    def select_clipboard_menu_option(self, item=None, label: str = "", labels=()):
        # ensure that only the matching menu item is selected,
        # (can be specified as a menuitem object, or using its label)
        # all the other menu items whose labels are specified will be made inactive
        # (we use a flag to prevent reentry)
        clipboardlog("select_clipboard_menu_option(%s, %s, %s) clipboard_change_pending=%s",
                     item, label, labels, self._clipboard_change_pending)
        if self._clipboard_change_pending:
            return None
        clipboard = self.get_menu("Clipboard")
        if not clipboard:
            log.error("Error: cannot locate Clipboard menu object!")
            return None
        all_items = [x for x in clipboard.get_submenu().get_children() if x.get_label() in labels]
        selected_items = [x for x in all_items if x == item] + [x for x in all_items if x.get_label() == label]
        if not selected_items:
            log.error("Error: cannot find any clipboard menu options to match '%s'", label)
            log.error(" all menu items: %s", csv(x.get_label() for x in all_items))
            log.error(" selected: %s", csv(x.get_label() for x in selected_items))
            return None
        self._clipboard_change_pending = True
        sel = selected_items[0]
        if not label:
            label = sel.get_label()
        for x in all_items:
            active = x.get_label() == label
            if x.get_active() != active:
                x.set_active(active)
        self._clipboard_change_pending = False
        return label

    def set_clipboard_menu(self, clipboard_menu) -> None:
        # find the menu item matching the current settings,
        # and select it
        try:
            ch = self.client.clipboard_helper
            rc_setting = "Clipboard"
            if len(ch._local_to_remote) == 1:
                rc_setting = tuple(ch._local_to_remote.values())[0]
            label = CLIPBOARD_NAME_TO_LABEL.get(rc_setting)
            clipboardlog(f"set_clipboard_menu(%s) setting={rc_setting!r}, {label=}", clipboard_menu)
            self.select_clipboard_menu_option(None, label, CLIPBOARD_LABELS)
        except RuntimeError:
            clipboardlog("failed to select remote clipboard option in menu", exc_info=True)
        direction = self.client.client_clipboard_direction
        direction_label = CLIPBOARD_DIRECTION_NAME_TO_LABEL.get(direction, "Disabled")
        clipboardlog("direction(%s)=%s", direction, direction_label)
        self.select_clipboard_menu_option(None, direction_label, CLIPBOARD_DIRECTION_LABELS)

    # these methods are called by the superclass,
    # but we don't have a quality or speed menu,
    # so override and ignore
    def set_qualitymenu(self, *args):
        pass  # no quality menu on MacOS

    def set_speedmenu(self, *args):
        pass  # no speed menu on MacOS

    def _get_keyboard(self):
        if not self.client or not self.client.keyboard_helper:
            return None
        return self.client.keyboard_helper.keyboard

    def make_swapkeysmenuitem(self):
        def swapkeys_toggled(*args):
            v = swapkeys_menuitem.get_active()
            keyboard = self._get_keyboard()
            log("swapkeys_toggled(%s) keyboard=%s, swap keys enabled=%s", args, keyboard, v)
            if keyboard:
                keyboard.swap_keys = v

        swapkeys_menuitem = self.checkitem("Control/Command Key Swap", swapkeys_toggled)

        def set_swapkeys_menuitem(*args):
            keyboard = self._get_keyboard()
            if keyboard:
                log("set_swapkeys_menuitem(%s) keyboard=%s, swap_keys=%s", args, keyboard, keyboard.swap_keys)
                swapkeys_menuitem.set_active(keyboard.swap_keys)
            else:
                log("set_swapkeys_menuitem(%s) no keyboard!", args)
                swapkeys_menuitem.set_sensitive(False)

        self.after_handshake(set_swapkeys_menuitem)
        return swapkeys_menuitem

    def make_invertmousewheelmenuitem(self):
        def invert_toggled(*args):
            v = mousewheel_menuitem.get_active()
            log("invert_toggled(%s) invert enabled=%s", args, v)
            if v:
                self.client.wheel_map[4] = 5
                self.client.wheel_map[5] = 4
            else:
                self.client.wheel_map[4] = 4
                self.client.wheel_map[5] = 5

        mousewheel_menuitem = self.checkitem("Invert Mouse Wheel", invert_toggled)
        mousewheel_menuitem.set_active(self.client.wheel_map.get(4) != 4)
        return mousewheel_menuitem

    def make_numlockmenuitem(self):
        def numlock_toggled(*args):
            v = self.numlock_menuitem.get_active()
            keyboard = self._get_keyboard()
            log("numlock_toggled(%s) menu active=%s", args, v)
            if keyboard:
                keyboard.num_lock_state = v

        self.numlock_menuitem = self.checkitem("Num Lock", cb=numlock_toggled)
        self.numlock_menuitem.set_active(True)

        def set_numlock_menuitem(*args):
            keyboard = self._get_keyboard()
            if keyboard:
                log("set_numlock_menuitem(%s) keyboard=%s, num_lock_state=%s", args, keyboard, keyboard.num_lock_state)
                self.numlock_menuitem.set_active(keyboard.num_lock_state)
            else:
                log("set_numlock_menuitem(%s) no keyboard!", args)
                self.numlock_menuitem.set_sensitive(False)

        self.after_handshake(set_numlock_menuitem)
        return self.numlock_menuitem

    def update_numlock(self, on):
        if self.numlock_menuitem:
            self.numlock_menuitem.set_active(on)

    def build_dock_menu(self):
        log("OSXMenuHelper.build_dock_menu()")
        self.dock_menu = self.make_menu()
        if SHOW_ABOUT_XPRA:
            self.dock_menu.add(self.menuitem("About Xpra", "information.png", None, about))
        self.dock_menu.show_all()
        return self.dock_menu

    # the code below is mostly duplicated from xpra/client/gtk2...

    def get_image(self, icon_name, size=None):
        try:
            pixbuf = get_icon(icon_name)
            log("get_image(%s, %s) pixbuf=%s", icon_name, size, pixbuf)
            if not pixbuf:
                return None
            if size:
                return scaled_image(pixbuf, size)
            return Gtk.Image.new_from_pixbuf(pixbuf)
        except Exception:
            log.error("get_image(%s, %s)", icon_name, size, exc_info=True)
            return None



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Final

from xpra.os_util import gi_import
from xpra.gtk.gobject import one_arg_signal
from xpra.gtk.error import xlog, xsync
from xpra.x11.gtk.damage import WindowDamageHandler
from xpra.x11.gtk.bindings import add_event_receiver, remove_event_receiver
from xpra.x11.gtk.world_window import get_world_window
from xpra.x11.bindings.ximage import XImageBindings
from xpra.x11.bindings.window import constants, X11WindowBindings
from xpra.log import Logger

log = Logger("x11", "window", "damage")

GObject = gi_import("GObject")

XImage = XImageBindings()
X11Window = X11WindowBindings()
X11Window.ensure_XComposite_support()

StructureNotifyMask: Final[int] = constants["StructureNotifyMask"]


class CompositeHelper(WindowDamageHandler, GObject.GObject):
    __gsignals__ = WindowDamageHandler.__common_gsignals__.copy()
    __gsignals__ |= {
        # emit:
        "contents-changed": one_arg_signal,
    }

    # This may raise XError.
    def __init__(self, xid: int):
        WindowDamageHandler.__init__(self, xid)
        GObject.GObject.__init__(self)
        self._listening_to: list[int] = []

    def __repr__(self):
        return f"CompositeHelper({self.xid:x})"

    def setup(self) -> None:
        X11Window.XCompositeRedirectWindow(self.xid)
        WindowDamageHandler.setup(self)

    def do_destroy(self) -> None:
        with xlog:
            X11Window.XCompositeUnredirectWindow(self.xid)
            WindowDamageHandler.do_destroy(self)

    def invalidate_pixmap(self) -> None:
        lt = self._listening_to
        if lt:
            self._listening_to = []
            self._cleanup_listening(lt)
        WindowDamageHandler.invalidate_pixmap(self)

    def _cleanup_listening(self, listening: list[int]) -> None:
        if listening:
            for w in listening:
                # Don't want to stop listening to our xid!:
                if w != self.xid:
                    remove_event_receiver(w, self)

    def _set_pixmap(self) -> None:
        # The tricky part here is that the pixmap returned by
        # NameWindowPixmap gets invalidated every time the window's
        # viewable state changes.  ("viewable" here is the X term that
        # means "mapped, and all ancestors are also mapped".)  But
        # there is no X event that will tell you when a window's
        # viewability changes!
        # Instead, we have to find all ancestors,
        # and watch all of them for unmap and reparent events.  But
        # what about races?  I hear you cry.  By doing things in the
        # exact order:
        #   1) select for StructureNotify
        #   2) QueryTree to get parent
        #   3) repeat 1 & 2 up to the root
        #   4) call NameWindowPixmap
        # we are safe.  (I think.)
        listening: list[int] = []
        e = None
        try:
            world = get_world_window()
            wxid = None
            if world:
                wxid = world.get_window().get_xid()
            with xsync:
                rxid = X11Window.get_root_xid()
            xid = X11Window.getParent(self.xid)
            while xid not in (0, rxid, wxid):
                # We have to use a lowlevel function to manipulate the
                # event selection here, because SubstructureRedirectMask
                # does not roundtrip through the GDK event mask
                # functions.  So if we used them, here, we would clobber
                # corral window selection masks, and those don't deserve
                # clobbering.  They are our friends!  X is driving me
                # slowly mad.
                parent = X11Window.getParent(xid)
                if not parent:
                    break
                X11Window.addXSelectInput(xid, StructureNotifyMask)
                add_event_receiver(xid, self, max_receivers=-1)
                listening.append(xid)
                xid = parent
            handle = XImage.get_xcomposite_pixmap(self.xid)
        except Exception:
            try:
                self._cleanup_listening(listening)
            except Exception:
                log(f"failed to cleanup listening for {listening}", exc_info=True)
            raise
        if handle is None:
            log("failed to name a window pixmap for %#x: %s", self.xid, e)
            self._cleanup_listening(listening)
        else:
            self._contents_handle = handle
            # Don't save the listening set until after
            # NameWindowPixmap has succeeded, to maintain our
            # invariant:
            self._listening_to = listening

    def do_x11_damage_event(self, event) -> None:
        event.x += self._border_width
        event.y += self._border_width
        self.emit("contents-changed", event)


GObject.type_register(CompositeHelper)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

import dbus.service

from xpra.dbus.helper import native_to_dbus
from xpra.dbus.common import init_system_bus, init_session_bus
from xpra.server.dbus.server import DBUS_Server_Base, INTERFACE, BUS_NAME
from xpra.log import Logger

log = Logger("dbus", "server")


class Proxy_DBUS_Server(DBUS_Server_Base):

    def __init__(self, server=None):
        if os.getuid() == 0:
            bus = init_system_bus()
        else:
            bus = init_session_bus()
        super().__init__(bus, server, BUS_NAME)

    @dbus.service.method(INTERFACE, in_signature='', out_signature='a{sv}')
    def GetInfo(self):
        # full info is available by calling get_info()
        i = self.server.get_minimal_server_info()
        self.log(".GetInfo()=%s", i)
        try:
            v = dbus.types.Dictionary((str(k), native_to_dbus(v)) for k, v in i.items())
            log("native_to_dbus(..)=%s", v)
        except Exception:
            log("GetInfo:gotinfo", exc_info=True)
            v = dbus.types.Dictionary({})
        return v
